# Check the following urls for more info about Pan-STARRS:
#
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-ImportantFITSimageformat,WCS,andflux-scalingnotes
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images#PS1Stackimages-Photometriccalibration
#
# For DES:
#
#     https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/processing
#
# For SDSS:
#
#     https://www.sdss.org/dr12/algorithms/fluxcal/#SDSStoAB
#     https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
#
# Some parts of this notebook are based on https://github.com/djones1040/PS1_surface_brightness/blob/master/Surface%20Brightness%20Tutorial.ipynb and codes from Lluís Galbany

import numpy as np
import pandas as pd
from pathlib import Path
from typing import Optional
import matplotlib.pyplot as plt

import aplpy
import sep
from photutils.aperture import aperture_photometry, CircularAperture
from photutils.utils import calc_total_error

from astropy.io import fits
from astropy.wcs import WCS
from astropy import units as u
from astropy.cosmology import FlatLambdaCDM, Cosmology

from hostphot._constants import workdir, font_family
from hostphot.processing.cleaning import remove_nan
from hostphot.photometry.dust import calc_extinction
from hostphot.utils import check_work_dir, suppress_stdout, store_input
from hostphot.photometry.image_utils import get_image_exptime
from hostphot.photometry.photometry_utils import magnitude_calculation
from hostphot.surveys_utils import (
    get_survey_filters,
    survey_pixel_units,
    check_filters_validity,
    check_survey_validity,
    survey_pixel_scale,
    bkg_surveys,
)

import warnings
from astropy.utils.exceptions import AstropyWarning

from photutils.background import Background2D, SExtractorBackground

plt.rcParams["mathtext.fontset"] = "cm"

# set initial cosmology
H0 = 70
Om0 = 0.3
__cosmo__ = FlatLambdaCDM(H0, Om0)
sep.set_sub_object_limit(1e4)


def choose_cosmology(cosmo: Cosmology) -> None:
    """Updates the cosmology used to calculate the aperture size.

    Parameters
    ----------
    cosmo: Cosmological model. E.g. :func:`FlatLambdaCDM(70, 0.3)`.
    """
    global __cosmo__
    __cosmo__ = cosmo


def calc_aperture_size(z: float, ap_radius: float) -> float:
    """Calculates the size of the aperture in arsec,
    for aperture photometry, given a physical size.

    Parameters
    ----------
    z: Redshift.
    ap_radius: Physical aperture size in kpc.

    Returns
    -------
    radius_arcsec: Aperture size in arcsec.
    """
    ap_radius = ap_radius * u.kpc
    # transverse separations
    transv_sep_per_arcmin = __cosmo__.kpc_proper_per_arcmin(z)
    transv_sep_per_arcsec = transv_sep_per_arcmin.to(u.kpc / u.arcsec)
    radius_arcsec = ap_radius / transv_sep_per_arcsec

    return radius_arcsec.value


def extract_aperture_flux(
    data: np.ndarray, bkg: np.ndarray, exptime: np.ndarray, px: float, py: float, radius: float
) -> tuple[float, float]:
    """Extracts aperture photometry of a single image.

    Parameters
    ----------
    data: Image data in a 2D numpy array.
    bkg: Background error of the images.
    exptime: Exposure time.
    px: x-axis pixel coordinate of the aperture center.
    py: y-axis pixel coordinate of the aperture center.
    radius: Aperture radius in pixels.

    Returns
    -------
    raw_flux: Aperture photometry ("raw" flux).
    raw_flux_err: Uncertainty on the aperture photometry.
    """
    error = calc_total_error(data, bkg, exptime)
     
    aperture = CircularAperture((px, py), r=radius)
    ap_results = aperture_photometry(data, aperture, error=error)
    raw_flux = ap_results["aperture_sum"][0]
    raw_flux_err = ap_results["aperture_sum_err"][0]

    return raw_flux, raw_flux_err


def plot_aperture(
    hdu: list[fits.ImageHDU],
    px: float,
    py: float,
    radius_pix: float,
    title: str,
    outfile: Optional[str | Path] = None,
) -> None:
    """Plots the aperture for the given parameters.

    Parameters
    ----------
    hdu: HDU image.
    px: x-axis center of the aperture in pixels.
    py: y-axis center of the aperture in pixels.
    radius_pix: Aperture radius in pixels.
    title: Title of the figure.
    outfile: If given, path where to save the output figure.
    """
    figure = plt.figure(figsize=(10, 10))
    title, label = title.split("|")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig = aplpy.FITSFigure(hdu, figure=figure)
    with suppress_stdout():
        fig.show_grayscale(stretch="arcsinh")

    # plot SN and aperture
    fig.show_markers(
        px,
        py,
        edgecolor="k",
        facecolor="aqua",
        marker="*",
        s=200,
        label="SN",
        coords_frame="pixel",
    )
    fig.show_circles(
        px, py, radius_pix, coords_frame="pixel", linewidth=3, edgecolor="r"
    )
    fig.show_lines([], [], color="r", lw=3, label=label)  # for the legend only

    # ticks
    fig.tick_labels.set_font(**{"family": font_family, "size": 18})
    fig.tick_labels.set_xformat("dd.dd")
    fig.tick_labels.set_yformat("dd.dd")
    fig.ticks.set_length(6)
    # ToDo: solve this deprecation warning (Aplpy should do it?)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig.axis_labels.set_font(**{"family": font_family, "size": 18})

    fig.set_title(title, **{"family": font_family, "size": 24})
    fig.set_theme("publication")
    fig.ax.legend(fancybox=True, framealpha=1, prop={"size": 20, "family": font_family})
    # output
    if outfile:
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(figure)
    else:
        plt.show()


def photometry(
    name: str,
    ra: float,
    dec: float,
    z: float,
    filt: str,
    survey: str,
    ap_radii: int | float | list = 1,
    ap_units: str = "kpc",
    bkg_sub: Optional[bool] = None,
    use_mask: bool = True,
    correct_extinction: bool = True,
    save_plots: bool = True,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, float]:
    """Calculates the local aperture photometry in a given radius.

    Parameters
    ----------
    name: Name of the object to find the path of the fits file.
    ra: Right Ascensions in degrees to center the aperture.
    dec: Declinations in degrees to center the aperture.
    z: Redshift of the object to estimate the physical calculate
        of the aperture.
    filt: Filter to use to load the fits file.
    survey: Survey to use for the zero-points and pixel scale.
    ap_radii: Aperture size.
    ap_units: Aperture size units. Either ``kpc`` or ``arcsec``.
    bkg_sub: If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    use_mask: If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    save_plots: If ``True``, the figure with the aperture is saved.

    Returns
    -------
    mags: Aperture magnitudes for the given aperture radii.
    mags_err: Aperture magnitude errors for the given aperture radii.
    fluxes: Aperture flux for the given aperture radii.
    fluxes_err: Aperture flux errors for the given aperture radii.
    zp: Zeropoint.
    """     
    # initial checks
    check_survey_validity(survey)
    check_work_dir(workdir)
    assert ap_units in ["kpc", "arcsec"], "not valid aperture size units"
    obj_dir = Path(workdir, name)
    if use_mask:
        suffix = "_masked"
    else:
        suffix = ""
    fits_file = obj_dir / survey / f"{survey}_{filt}{suffix}.fits"

    # load image information
    hdu = fits.open(fits_file)
    hdu = remove_nan(hdu)
    header = hdu[0].header
    data = hdu[0].data
    exptime = get_image_exptime(header, survey)
    pixel_units = survey_pixel_units(survey, filt)
    if pixel_units == "counts":
        _exptime = 1
    else:
        _exptime = exptime
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        wcs = WCS(header, naxis=2)

    data = data.astype(np.float64)
    #box_size = int(0.1 * np.sqrt(data.size))
    #bkg = SExtractorBackground(sigma_clip=None)
    #bkg = Background2D(data, box_size=box_size, bkg_estimator=bkg)
    bkg = sep.Background(data)
    # background subtraction, if needed
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
        #data_sub = np.copy(data - bkg.background)
    else:
        data_sub = np.copy(data)

    # turn float into a list for consistency
    if isinstance(ap_radii, (float, int)):
        ap_radii = [ap_radii]

    pixel_scale = survey_pixel_scale(survey, filt)
    if survey == "UKIDSS" and filt == "J":
        if "LAS" in header["PROJECT"].split("/")[-1]:
            # if the J comes from the LAS, then the observations were micro-stepped,
            # so the pixel size is smaller.
            nustep = header["NUSTEP"]
            pixel_scale /= nustep / 2  # divided by 2 for the two dimensions (x and y)

    mags, mags_err = [], []
    fluxes, fluxes_err = [], []
    px, py = wcs.wcs_world2pix(ra, dec, 0)  # object's coordinates

    for ap_radius in ap_radii:
        # aperture photometry
        if ap_units == "kpc":
            radius_arcsec = calc_aperture_size(z, ap_radius)
        else:
            radius_arcsec = ap_radius
        radius_pix = radius_arcsec / pixel_scale
        flux, flux_err = extract_aperture_flux(data_sub, bkg.rms(), _exptime, px, py, radius_pix)
        #flux, flux_err = extract_aperture_flux(data_sub, bkg.background_rms, _exptime, px, py, radius_pix)

        ap_area = np.pi * (radius_pix**2)
        mag, mag_err, flux, flux_err, zp = magnitude_calculation(
            flux,
            flux_err,
            survey,
            filt,
            ap_area,
            header,
            bkg.globalrms,
            #np.median(bkg.background_rms),
        )

        if correct_extinction is True:
            A_ext = calc_extinction(filt, survey, ra, dec)
            mag -= A_ext
            flux *= 10 ** (0.4 * A_ext)

        mags.append(mag)
        mags_err.append(mag_err)
        fluxes.append(flux)
        fluxes_err.append(flux_err)

        if save_plots:
            outfile = obj_dir / survey / f"local_{survey}_{filt}_{ap_radius}{ap_units}.jpg"
            filt_ = filt.replace("_", "-")
            if ap_units == "kpc":
                title = rf"{name}: {survey}-${filt_}$|r$={ap_radius}$ {ap_units} @ $z={z}$"
            else:
                title = rf"{name}: {survey}-${filt_}$|r$={ap_radius}$ {ap_units}"
            plot_aperture(hdu, px, py, radius_pix, title, outfile)
    hdu.close()

    return mags, mags_err, fluxes, fluxes_err, zp


def multi_band_phot(
    name: str,
    ra: float,
    dec: float,
    z: float,
    filters: Optional[str | list] = None,
    survey: str = "PanSTARRS",
    ap_radii: int | float | list = 1,
    ap_units: str = "kpc",
    bkg_sub: Optional[bool] = None,
    use_mask: bool = True,
    correct_extinction: bool = True,
    save_plots: bool = True,
    save_results: bool = True,
    raise_exception: bool = True,
    save_input: bool = True,
) -> pd.DataFrame:
    """Calculates the local aperture photometry for multiple filters.

    Parameters
    ----------
    name: Name of the object to find the path of the fits file.
    ra: Right Ascensions in degrees to center the aperture.
    dec: Declinations in degrees to center the aperture.
    z: Redshift of the object to estimate the physical calculate
        of the aperture.
    filters: Filters to use to load the fits files. If ``None`` use all
        the filters of the given survey.
    survey: Survey to use for the zero-points and pixel scale.
    ap_radii: Aperture size.
    ap_units: Aperture size units. Either ``kpc`` or ``arcsec``.
    bkg_sub: If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    use_mask: If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    save_plots: If ``True``, the figure with the aperture is saved.
    save_results: If ``True``, the magnitudes are saved into a csv file.
    raise_exception: If ``True``, an exception is raised if the photometry fails for any filter.
    save_input: Whether to save the input parameters.

    Returns
    -------
    phot_df: DataFrame with the object's photometry and other info.

    Examples
    --------
    >>> from hostphot.photometry import local_photometry as lp
    >>> ap_radii = [3, 4]  # aperture radii in units of kpc
    >>> ra, dec =  308.22579, 9.92853 # coords of SN2004eo
    >>> z = 0.0157  # redshift
    >>> survey = 'PanSTARRS'
    >>> results = lp.multi_band_phot(name, 
                                     ra, 
                                     dec, 
                                     z,
                                     survey=survey, 
                                     ap_radii=ap_radii, 
                                     use_mask=True, 
                                     save_plots=True)
    """
    input_params = locals()  # dictionary
    # initial checks
    check_survey_validity(survey)
    if filters is None:
        if survey in ["HST", "JWST"]:
            raise ValueError(f"For {survey}, the filter needs to be specified!")
        filters = get_survey_filters(survey)
    else:
        check_filters_validity(filters, survey)
    #if survey in ["HST", "JWST"]:
    #     filters = [filters]
    assert ap_units in ["kpc", "arcsec"], "not valid aperture size units"

    # save input parameters
    if save_input is True:
        inputs_file = Path(workdir, name, survey, "input_local_photometry.csv")
        store_input(input_params, inputs_file)

    # turn int/float into a list for consistency
    if isinstance(ap_radii, (float, int)):
        ap_radii = [ap_radii]

    results_dict = {
        "name": name,
        "ra": ra,
        "dec": dec,
        "redshift": z,
        "survey": survey,
        "ap_units": ap_units,
    }
    # calculate photometry
    for filt in filters:
        try:
            mags, mags_err, fluxes, fluxes_err, zp = photometry(
                name,
                ra,
                dec,
                z,
                filt,
                survey,
                ap_radii,
                ap_units,
                bkg_sub,
                use_mask,
                correct_extinction,
                save_plots,
            )
            for radius, mag, mag_err, flux, flux_err in zip(
                ap_radii, mags, mags_err, fluxes, fluxes_err
            ):
                results_dict[f"{filt}_{radius}"] = mag
                results_dict[f"{filt}_{radius}_err"] = mag_err
                results_dict[f"{filt}_{radius}_flux"] = flux
                results_dict[f"{filt}_{radius}_flux_err"] = flux_err
            results_dict[f"{filt}_zeropoint"] = zp
        except Exception as exc:
            if raise_exception is True:
                raise Exception(exc)
            else:
                for radius in ap_radii:
                    results_dict[f"{filt}_{radius}"] = np.nan
                    results_dict[f"{filt}_{radius}_err"] = np.nan
                    results_dict[f"{filt}_{radius}_flux"] = np.nan
                    results_dict[f"{filt}_{radius}_flux_err"] = np.nan
                results_dict[f"{filt}_zeropoint"] = np.nan
    # store photometry
    phot_df = pd.DataFrame({key: [val] for key, val in results_dict.items()})
    if save_results is True:
        outfile = Path(workdir, name, survey, "local_photometry.csv")
        phot_df.to_csv(outfile, index=False)

    return phot_df



import numpy as np
from pathlib import Path

import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astropy.coordinates import SkyCoord

from astroquery.esa.hubble import ESAHubble  # HST

from hostphot._constants import workdir
from hostphot.utils import check_work_dir, suppress_stdout
from hostphot.surveys_utils import check_HST_filters, survey_pixel_scale

import warnings
from astropy.utils.exceptions import AstropyWarning


def update_HST_header(hdu: fits.hdu.ImageHDU) -> None:
    """Updates the HST image header with the necessary keywords.

    Parameters
    ----------
    hdu : Header Data Unit.
        HST FITS image.
    """
    # Drizzlepac images have the info on hdu[0]
    # MAST-archive images have the info on hdu[1]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        if "PHOTPLAM" not in hdu[0].header:
            # MAST image: move things to hdu[0] to homogenise
            img_wcs = WCS(hdu[1].header)
            hdu[0].header.update(img_wcs.to_header())
            hdu[0].header["PHOTFLAM"] = hdu[1].header["PHOTFLAM"]
            hdu[0].header["PHOTPLAM"] = hdu[1].header["PHOTPLAM"]
            hdu[0].data = hdu[1].data
    # add zeropoints
    # https://www.stsci.edu/hst/instrumentation/acs/data-analysis/zeropoints
    photflam = hdu[0].header["PHOTFLAM"]
    photplam = hdu[0].header["PHOTPLAM"]
    hdu[0].header["MAGZP"] = (
        -2.5 * np.log10(photflam) - 5 * np.log10(photplam) - 2.408
    )

def set_HST_image(file: str, filt: str, name: str) -> None:
    """Moves a previously downloaded HST image into the work directory.

    The image's header is updated with the necessary keywords to obtain
    photometry and is also moved under the objects directory inside the
    work directory.

    HST images take very long to download, so the user might prefer to
    download the images manually and then use this function to include
    the image into the workflow.

    Parameters
    ----------
    file: HST image to use.
    filt: HST filter, e.g. ``WFC3_UVIS_F275W``.
    name: Object's name under which the image is saved.
    """
    # check output directory
    check_work_dir(workdir)
    check_HST_filters(filt)
    obj_dir = Path(workdir, name, "HST")
    if obj_dir.is_dir() is False:
        obj_dir.mkdir(parents=True, exist_ok=True)
    # update file and save file
    hdu = fits.open(file)
    update_HST_header(hdu)
    outfile = obj_dir / f"HST_{filt}.fits"
    hdu.writeto(outfile, overwrite=True)

def get_HST_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                        filters: list = ["WFC3_UVIS_F225W"]) -> list[fits.ImageHDU]:
    """Downloads a set of HST fits images for a given set
    of coordinates and filters using astroquery.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use, e.g. ``WFC3_UVIS_F225W``.

    Return
    ------
    hdu_list: List with fits image for the given filter. ``None`` is returned if no image is found.
    """
    esahubble = ESAHubble()
    esahubble.get_status_messages()
    
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)
    size_arcsec = size_arcsec.value
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
    
    # query observations at the given coordinates
    with suppress_stdout():
        result = esahubble.cone_search_criteria(
            radius=3,
            coordinates=coords,
            calibration_level="PRODUCT",
            data_product_type="image",
            #instrument_name=instrument,
            #filters=filt,
            async_job=True,
        )
    results_df = result.to_pandas()
    
    hdu_list = []
    for filt in filters:
        check_HST_filters(filt)
        # separate the instrument name from the actual filter
        split_filt = filt.split("_")
        if len(split_filt) == 2:
            filt = split_filt[-1]
            instrument = split_filt[0]
        elif len(split_filt) == 3:
            filt = split_filt[-1]
            instrument = f"{split_filt[0]}/{split_filt[1]}"
        else:
            raise ValueError(f"Incorrect filter name: {filt}")
        # filter by filter and instrument
        obs_df = results_df[results_df["filter"] == filt]
        obs_df = obs_df[obs_df.instrument_name == instrument]
        # get only exposures shorter than one hour
        obs_df = obs_df[obs_df.exposure_duration < 3600]
        obs_df.sort_values(
            by=["exposure_duration"], ascending=False, inplace=True
        )

        # start download 
        filename = f"HST_tmp_{ra}_{dec}"  # the extension is added below
        for obs_id in obs_df.observation_id:
            try:
                esahubble.download_product(
                    observation_id=obs_id,
                    product_type="SCIENCE",
                    calibration_level="PRODUCT",
                    filename=filename,
                )
                break
            except:
                pass
        temp_file = Path(f"{filename}.fits.gz")
        if temp_file.is_file() is False:
            hdu_list.append(None)
            continue
        hdu = fits.open(temp_file)
        # remove the temporary files
        temp_file.unlink()
        # add necessary information to the header
        update_HST_header(hdu)
        hdu_list.append(hdu)
    # HST images are large so need to be trimmed
    for hdu, filt in zip(hdu_list, filters):
        if hdu is None:
            continue
        pixel_scale = survey_pixel_scale("HST", filt)  # same pixel scale for all filters
        size_pixels = int(size_arcsec / pixel_scale)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            img_wcs = WCS(hdu[0].header)
        trimmed_data = Cutout2D(hdu[0].data, coords, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())
    return hdu_list


import tarfile
import requests
import numpy as np
from pathlib import Path

import extinction
from sfdmap2 import sfdmap
from astropy.coordinates import SkyCoord
from astropy import units as u

import hostphot
from hostphot.photometry.photometry_utils import integrate_filter, extract_filter

import warnings


def _download_dustmaps():
    """Downloads the dust maps for extinction calculation if they are not found
    locally.
    """
    mapsdir = Path(hostphot.__path__[0])

    # check if files already exist locally
    dust_files = [
        mapsdir.joinpath("sfddata-master", rf"SFD_dust_4096_{sky}gp.fits")
        for sky in ["n", "s"]
    ]
    mask_files = [
        mapsdir.joinpath("sfddata-master", rf"SFD_mask_4096_{sky}gp.fits")
        for sky in ["n", "s"]
    ]
    maps_files = dust_files + mask_files
    existing_files = [file.is_file() for file in maps_files]

    if not all(existing_files) == True:
        # download dust maps
        sfdmaps_url = "https://github.com/kbarbary/sfddata/archive/master.tar.gz"
        response = requests.get(sfdmaps_url)

        master_tar = Path("master.tar.gz")
        with open(master_tar, "wb") as file:
            file.write(response.content)

        # extract tar file under mapsdir directory
        tar = tarfile.open(master_tar)
        tar.extractall(mapsdir)
        tar.close()
        master_tar.unlink()  # remove file


def deredden(
    wave: np.ndarray,
    flux: np.ndarray,
    ra: float,
    dec: float,
    scaling: float = 0.86,
    reddening_law: str = "fitzpatrick99",
    r_v: float = 3.1,
) -> np.ndarray:
    """Dereddens the given spectrum, given a right ascension and declination or :math:`E(B-V)`.

    Parameters
    ----------
    wave: Wavelength values.
    flux: Flux density values.
    ra:  Right ascension in degrees.
    dec: Declination in degrees.
    scaling: Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989),
        ``odonnell94`` (O'Donnell 1994), ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00``
        (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with :math:`R_V` = 3.1.)
    r_v: Total-to-selective extinction ratio (:math:`R_V`)

    Returns
    -------
    deredden_flux : array
        Deredden flux values.
    """
    _download_dustmaps()

    hostphot_path = Path(hostphot.__path__[0])
    dustmaps_dir = hostphot_path / "sfddata-master"

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        m = sfdmap.SFDMap(mapdir=dustmaps_dir, scaling=scaling)
        ebv = m.ebv(ra, dec)  # RA and DEC in degrees
        a_v = r_v * ebv

    rl_list = ["ccm89", "odonnell94", "fitzpatrick99", "calzetti00", "fm07"]
    assert (
        reddening_law in rl_list
    ), f"Choose one of the available reddening laws: {rl_list}"

    if reddening_law == "ccm89":
        ext = extinction.ccm89(wave, a_v, r_v)
    elif reddening_law == "odonnell94":
        ext = extinction.odonnell94(wave, a_v, r_v)
    elif reddening_law == "fitzpatrick99":
        ext = extinction.fitzpatrick99(wave, a_v, r_v)
    elif reddening_law == "calzetti00":
        ext = extinction.calzetti00(wave, a_v, r_v)
    elif reddening_law == "fm07":
        ext = extinction.fm07(wave, a_v)

    deredden_flux = extinction.remove(ext, flux)

    return deredden_flux


def calc_extinction(
    filt: str,
    survey: str,
    ra: float,
    dec: float,
    scaling: float = 0.86,
    reddening_law: str = "fitzpatrick99",
    r_v: float = 3.1,
):
    """Calculates the extinction for a given filter, right ascension and declination
    or `E(B-V)`.

    Parameters
    ----------
    filter_wave: Filter's wavelength range.
    filter_response: Filter's response function.
    ra: Right ascension.
    dec: Declinationin degrees.
    scaling: Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989),
        ``odonnell94`` (O'Donnell 1994), ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00``
        (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with :math:`R_V` = 3.1.)
    r_v: Total-to-selective extinction ratio (:math:`R_V`)

    Returns
    -------
    A_ext: Extinction value in magnitudes.
    """
    if survey == "LegacySurvey":
        # https://datalab.noirlab.edu/ls/bass.php
        # declination above ~32 and above the galactic plane
        gal_coords = SkyCoord(ra=ra * u.degree, dec=dec * u.degree, frame="icrs")
        if (dec > 32.375) and (gal_coords.b.value > 0):
            version = "BASS+MzLS"
        else:
            version = "DECam"
    else:
        version = None

    # extract transmission function for the given filter+survey
    filter_wave, filter_response = extract_filter(filt, survey, version)

    # calculate extinction
    flux = 100
    dereddened_flux = deredden(filter_wave, flux, ra, dec, scaling, reddening_law, r_v)

    f1 = integrate_filter(filter_wave, flux, filter_wave, filter_response)
    f2 = integrate_filter(filter_wave, dereddened_flux, filter_wave, filter_response)
    A_ext = -2.5 * np.log10(f1 / f2)

    return A_ext



import numpy as np
import pandas as pd
from pathlib import Path
from matplotlib import ticker
import matplotlib.pyplot as plt
from typing import Optional

import hostphot
from hostphot._constants import workdir, font_family
from hostphot.surveys_utils import get_survey_filters, filters_file, load_yml, extract_filter

path = Path(hostphot.__path__[0])
filters_config = load_yml(filters_file)


def get_eff_wave(filt: str, survey: str, version: str=None) -> float:
    """Obtains the effective wavelength of a filter.

    Parameters
    ----------
    filt: Filter name.
    survey: Survey name.
    version: Survey version.

    Returns
    -------
    eff_wave: Effective wavelength in angstroms.
    """
    wave, trans = extract_filter(filt, survey, version)
    eff_wave = np.sum(wave * trans) / np.sum(trans)

    return eff_wave


def plot_sed(
    name: str,
    phot_type: str = "global",
    z: Optional[float] = None,
    radius: str | int | float = None,
    ap_units: str = "kpc",
    include: list = None,
    exclude: list = None,
    save_plot: bool = True,
    outfile: str | Path = None,
    plot_flux: bool = False,
) -> None:
    """Plots the SED of an object.

    The SED will depend on the available photometry.

    Parameters
    ----------
    name: Object name.
    phot_type: Type of photometry: ``global`` or ``local``. 
    z: Redshift of the object, by default ``None``. If given, corrects
        for time dilation.
    radius : Radius for the local photometry, by default ``None``.
    ap_units: Aperture untis for the local photometry.
    include: List of surveys to include in the plot. Cannot be given together
        with '``exclude``.
    exclude: List of surveys to exclude from the plot. Cannot be given together
        with '``include``.
    save_plot: Whether to save the SED plot.
    outfile: If give, the plot is saved with this name instead of the default ones.

    Raises
    ------
    ValueError
        The photometry type should be either ``global`` or ``local``.
    """
    if phot_type == "local":
        assert radius is not None, "radius must be given with local photometry"

    obj_path = Path(workdir, name)
    phot_files = [file for file in obj_path.rglob(f"*/{phot_type}_photometry.csv")]

    if include is not None and exclude is not None:
        raise ValueError("'inlcude' cannot be given together with 'exclude'!")

    # include or exclude some surveys
    if include is not None:
        include_files = []
        for file in phot_files:
            for pattern in include:
                if pattern in str(file):
                    include_files.append(file)
                    break
        phot_files = include_files

    if exclude is not None:
        include_files = []
        for file in phot_files:
            skip = False
            for pattern in exclude:
                if pattern in str(file):
                    skip = True
            if skip is False:
                include_files.append(file)
        phot_files = include_files

    if len(phot_files) == 0:
        print(f"There is no photometry for {name}!")
        return None

    # start plotting
    fig, ax = plt.subplots(figsize=(12, 6))
    for file in phot_files:
        phot_df = pd.read_csv(file)
        survey = file.parts[-2]
        filters = get_survey_filters(survey)

        # Vega to AB
        ab_offset = {filt: filters_config[survey][filt]["AB_offset"] 
                     for filt in filters}

        # adapt name for local photometry
        if phot_type == "local":
            ap_units = phot_df.ap_units.values[0]
            title = fr"{name} - {phot_type} SED (r $= {radius}$ {ap_units})"
            ext = f"_{radius}"
        elif phot_type == "global":
            title = f"{name} - {phot_type} SED"
            ext = ""
        else:
            raise ValueError(f"Invalid photometry type: {phot_type}")

        
        waves, phot, phot_err = [], [], []
        valid_filters = []
        for filt in filters:
            filt_str = filt + ext
            filt_err_str = filt + ext + "_err"
            if filt_str not in phot_df.columns:
                continue
            
            # get photometry
            version = None
            if survey == "LegacySurvey":
                version = "DECam"
            wave = get_eff_wave(filt, survey, version)
            mag = phot_df[filt_str].values[0] + ab_offset[filt]
            mag_err = phot_df[filt_err_str].values[0]
            waves.append(wave)
            phot.append(mag)
            phot_err.append(mag_err)
            valid_filters.append(filt)

        if z is not None:
            # correct for time dilation
            waves = np.array(waves) / (1 + z)
            phot = np.array(phot) #- 2.5 * np.log10((1 + z))
            xlabel = r"Rest Wavelength ($\AA$)"
            if ")" in title:
                title = title.replace(")", fr" @ $z={z}$)")
            else:
                title = title + fr" ($z={z}$)"
        else:
            waves = np.array(waves)
            phot = np.array(phot)
            xlabel = r"Observed Wavelength ($\AA$)"
        phot_err = np.array(phot_err)

        # NaN mask
        mask = ~np.isnan(phot) * phot_err > 0
        waves = waves[mask]
        phot = phot[mask]
        phot_err = phot_err[mask]
        valid_filters = np.array(valid_filters)[mask]
        label = f'{survey}\n{", ".join(filt for filt in valid_filters)}'

        if plot_flux is False:
            lims = phot_err > 1  # upper limits; inverted for magnitude plots
            phot_err[lims] = 0.3  # for visualization
            ax.errorbar(
                waves,
                phot,
                yerr=phot_err,
                marker="o",
                ms=10,
                ls="dashed",
                lw=3,
                #c=colours[survey],
                mec="k",
                capsize=6,
                label=label,
                lolims=lims,
            )
        else:
            flux = 10 ** (-0.4 * (phot - 23.9))
            flux_err = np.abs(flux * 0.4 * np.log(10) * phot_err)
            ax.errorbar(
                waves,
                flux,
                yerr=flux_err,
                marker="o",
                ms=10,
                ls="dashed",
                lw=3,
                #c=colours[survey],
                mec="k",
                capsize=6,
                label=label,
            )

    if len(phot_files) > 4:
        ncol = 2
    else:
        ncol = 1

    if plot_flux is False:
        ax.invert_yaxis()  # for magnitude plot
        ax.set_ylabel("AB Magnitude", fontsize=24, font=font_family)
    else:
        ax.set_ylabel("Flux (mJy)", fontsize=24, font=font_family)
    ax.set_xlabel(xlabel, fontsize=24, font=font_family)
    ax.set_title(title, fontsize=28, fontfamily=font_family)

    for label in ax.get_xticklabels():
        label.set_fontproperties(font_family)
    for label in ax.get_yticklabels():
        label.set_fontproperties(font_family)
    ax.tick_params(labelsize=20)

    ax.legend(
        ncol=ncol, fancybox=True, framealpha=0.7, prop={"size": 14, "family": font_family}
    )
    ax.set_xscale("log")
    # format ticks
    """
    ticks = np.array([2e3, 4e3, 9e3, 2e4, 4e4, 9e4, 2e5, 4e5, 9e5])
    start, end = ax.get_xlim()
    mask = (ticks >= start) & (ticks <= end)
    ax.set_xticks(ticks[mask])
    formatter = ticker.ScalarFormatter(useMathText=True)
    ax.get_xaxis().set_major_formatter(formatter)
    """

    if save_plot is True:
        if outfile is None:
            obj_dir = Path(workdir, name)
            basename = f"sed_{phot_type}.jpg"
            outfile = obj_dir / basename
        plt.tight_layout()
        plt.savefig(outfile)
    plt.show()



import numpy as np
from typing import Optional

from astropy.io import fits
from astropy import units as u
from astropy.table import Table

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

def query_ps1(ra: float, dec: float, size: float | u.Quantity = 3, filters: Optional[str] = None) -> Table:
    """Query 'ps1filenames.py' service to get a list of images

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    table: Astropy table with the results.
    """
    # check filters
    survey = "PanSTARRS"
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)
    if isinstance(filters, list):
        filters = "".join(filt for filt in filters)
    # get size in pixels (all filters have the same size)
    pixel_scale = survey_pixel_scale(survey, "g")
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)
    # download table
    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    url = (
        f"{service}?ra={ra}&dec={dec}&size={size_pixels}&format=fits&"
        f"filters={filters}"
    )

    return Table.read(url, format="ascii")

def get_PS1_urls(ra: float, dec: float, size: float | u.Quantity = 3, filters: Optional[str] = None) -> list[str]:
    """Get URLs for images obtained with :func:`query_ps1()`.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    url_list: List of URLs for the fits images.
    """
    # check filters
    survey = "PanSTARRS"
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)
    # get size in pixels (all filters have the same size)
    pixel_scale = survey_pixel_scale(survey, "g")
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)
    # get PS1 table with ULRs
    table = query_ps1(ra, dec, size=size, filters=filters)
    # sort filters from blue to red
    flist = ["grizy".find(x) for x in table["filter"]]
    table = table[np.argsort(flist)]
    # get list of URLs for the images
    url = (
        "https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
        f"ra={ra}&dec={dec}&size={size_pixels}&format=fits"
    )
    base_url = url + "&red="
    url_list = []
    for filename in table["filename"]:
        url_list.append(base_url + filename)
    return url_list

def get_PanSTARRS_images(ra: float, dec: float, size: float | u.Quantity = 3, filters: Optional[str] = None) -> fits.HDUList:
    """Gets PanSTARRS fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    fits_files: List of fits images.
    """
    # check filters
    survey = "PanSTARRS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    # get URLs
    fits_url = get_PS1_urls(ra, dec, size, filters)
    # download FITS images
    hdu_list = []
    for url, filt in zip(fits_url, filters):
        hdu = fits.open(url)
        hdu_list.append(hdu)
    return hdu_list



import requests
from typing import Optional

import astropy.units as u
from astropy.io import fits

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

def get_Spitzer_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                        filters: Optional[str] = None) -> list[fits.ImageHDU]:
    """Gets Spitzer fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``IRAC.1, IRAC.2, IRAC.3, IRAC.4, MIPS.1``.

    Returns
    -------
    fits_files: List of fits images.
    """
    survey = "Spitzer"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    pixel_scale = survey_pixel_scale(survey, "IRAC.1")  # IRAC resolution
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    base_url = "https://irsa.ipac.caltech.edu/cgi-bin/Cutouts/nph-cutouts?"
    locstr = f"{str(ra)}+{str(dec)}+eq"

    mission_dict = {
        "mission": "SEIP",
        "min_size": "18",
        "max_size": "1800",
        "units": "arcsec",
        "locstr": locstr,
        "sizeX": size_pixels,
        "ntable_cutouts": "1",
        "cutouttbl1": "science",
        "mode": "PI",  # XML output file
    }

    # check image size
    if size_pixels < int(mission_dict["min_size"]):
        mission_dict["sizeX"] = int(mission_dict["min_size"])
        print(
            f'Image cannot be smaller than {mission_dict["min_size"]} arcsec (using this value)'
        )
    elif size_pixels > int(mission_dict["max_size"]):
        mission_dict["sizeX"] = int(mission_dict["max_size"])
        print(
            f'Image cannot be larger than {mission_dict["max_size"]} arcsec (using this value)'
        )

    mission_url = "&".join(
        [f"{key}={val}" for key, val in mission_dict.items()]
    )
    url = base_url + mission_url
    response = requests.get(url)
    list_text = response.text.split()

    # find the fits file for each fitler (None if not found)
    files_dict = {filt: None for filt in filters}
    for filt in filters:
        for line in list_text:
            if filt in line and line.endswith(".mosaic.fits"):
                files_dict[filt] = line
                # pick the first image and move to the next filter
                break

    hdu_list = []
    for filt, file in files_dict.items():
        if file is not None:
            hdu = fits.open(file)
            hdu[0].header["MAGZP"] = hdu[0].header["ZPAB"]
            hdu_list.append(hdu)
        else:
            hdu_list.append(None)
    return hdu_list


import unittest
import numpy as np
from hostphot.cutouts import download_images
from hostphot.photometry import global_photometry as gp


class TestHostPhot(unittest.TestCase):
    def test_global_phot(self):
        sn_name = "2002fk"
        host_ra = 50.527333
        host_dec = -15.400056
        survey = "PanSTARRS"

        download_images(sn_name, host_ra, host_dec, survey=survey, overwrite=False)
        phot = gp.multi_band_phot(
            sn_name,
            host_ra,
            host_dec,
            survey=survey,
            use_mask=False,
            common_aperture=False, 
            optimize_kronrad=True,
            save_plots=False,
            raise_exception=True,
        )
        mags = [phot[filt][0] for filt in "griz"]
        # griz reference magnitudes compared to Blast
        ref_mags = [11.684, 11.353, 11.215, 11.073]

        err_msg = (
            "Large difference between calculated and reference magnitudes"
        )
        np.testing.assert_allclose(mags, ref_mags, rtol=0.05, err_msg=err_msg)


if __name__ == "__main__":
    unittest.main()



# Check the following urls for more info about Pan-STARRS:
#
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-ImportantFITSimageformat,WCS,andflux-scalingnotes
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images#PS1Stackimages-Photometriccalibration
#
# For DES:
#
#     https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/processing
#
# For SDSS:
#
#     https://www.sdss.org/dr12/algorithms/fluxcal/#SDSStoAB
#     https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
#
# Some parts of this notebook are based on https://github.com/djones1040/PS1_surface_brightness/blob/master/Surface%20Brightness%20Tutorial.ipynb and codes from Lluís Galbany

import numpy as np
import pandas as pd
from pathlib import Path
from typing import Optional

import sep
from astropy.wcs import WCS
from astropy.io import fits

from photutils.utils import calc_total_error
from photutils.aperture import aperture_photometry, EllipticalAperture

from hostphot._constants import workdir
from hostphot.processing.objects_detection import extract_objects, plot_detected_objects
from hostphot.processing.cleaning import remove_nan
from hostphot.photometry.dust import calc_extinction
from hostphot.utils import check_work_dir, store_input
from hostphot.photometry.photometry_utils import magnitude_calculation
from hostphot.photometry.image_utils import adapt_aperture, get_image_exptime
from hostphot.surveys_utils import (
    get_survey_filters,
    survey_pixel_units,
    check_filters_validity,
    check_survey_validity,
    flipped_surveys,
    bkg_surveys,
)

import warnings
from astropy.utils.exceptions import AstropyWarning

sep.set_sub_object_limit(1e4)

def kron_flux(
    data: np.ndarray,
    bkg: float,
    exptime: float,
    objects: np.ndarray,
    kronrad: float,
    scale: float,
) -> tuple[float, float]:
    """Calculates the Kron flux.

    Parameters
    ----------
    data: Data of an image.
    bkg: Background error of the images.
    exptime: Exposure time.
    objects: Objects detected with `sep.extract()`.
    kronrad: Kron radius.
    scale: Scale of the Kron radius.

    Returns
    -------
    flux: Kron flux.
    flux_err: Kron flux error.
    """
    # theta must be in the range [-pi/2, pi/2] for sep.sum_ellipse()
    if objects["theta"] > np.pi / 2:
        objects["theta"] -= np.pi
    elif objects["theta"] < -np.pi / 2:
        objects["theta"] += np.pi
        
    error = calc_total_error(data, bkg, exptime)
    positions = np.array([objects["x"], objects["y"]]).T
    aperture = EllipticalAperture(positions, 
                                    objects["a"][0] * scale * kronrad, 
                                    objects["b"][0] * scale * kronrad,
                                    objects["theta"][0]
                                    )
    phot_table = aperture_photometry(data, aperture, error=error)
    flux = phot_table["aperture_sum"].value[0]
    flux_err = phot_table["aperture_sum_err"].value[0]

    return flux, flux_err


def optimize_aperture(
    data: np.ndarray, bkg: float, exptime: float, objects: np.ndarray, eps: float = 0.001
) -> tuple[float, float, float, float]:
    """Optimizes the Kron flux by iteration over different scales.

    The stop condition is met when the fractional change in flux between iterations
    is less that ``eps``.

    Parameters
    ----------
    data: Data of an image.
    bkg: Background error of the images.
    exptime: Exposure time.
    objects: Objects detected with :func:`sep.extract()`.
    eps: Minimum percent change in flux allowed between iterations.

    Returns
    -------
    opt_flux: Optimized Kron flux.
    opt_flux_err: Optimized Kron flux error.
    kronrad: Kron radius.
    opt_scale: Optimized scale for the Kron radius.
    """
    kronrad, _ = sep.kron_radius(
        data,
        objects["x"],
        objects["y"],
        objects["a"],
        objects["b"],
        objects["theta"],
        r=6.0,  # following sep docs
    )
    kronrad = kronrad[0]

    opt_flux = 0.0
    # iterate over scales
    scales = np.arange(1, 10, 0.01)
    for scale in scales:
        flux, flux_err = kron_flux(data, bkg, exptime, objects, kronrad, scale)
        calc_eps = np.abs(opt_flux - flux) / flux
        opt_flux, opt_flux_err = flux, flux_err
        opt_scale = scale
        if calc_eps < eps:
            opt_scale = scale
            break
        elif np.isnan(calc_eps):
            opt_scale = scales[-2]
            warnings.warn("Warning: the aperture might not fit in the image!")
            break

    return opt_flux, opt_flux_err, kronrad, opt_scale


def extract_aperture(
    name: str,
    host_ra: float,
    host_dec: float,
    filt: str,
    survey: str,
    ra: float = None,
    dec: float = None,
    bkg_sub: bool = False,
    threshold: float = 10,
    use_mask: bool = True,
    optimize_kronrad: bool = True,
    eps: float = 0.0001,
    gal_dist_thresh: float = -1,
    deblend_cont: float = 0.005,
    save_plots: bool = True,
    save_aperture_params: bool = True,
) -> tuple[np.ndarray, WCS, float, float, bool]:
    """Calculates the aperture for a given galaxy.

    Parameters
    ----------
    name: Name of the object to find the path of the fits file.
    host_ra: Host-galaxy right ascension of the galaxy in degrees.
    host_dec: Host-galaxy declination of the galaxy in degrees.
    filt: Filter to use to load the fits file. List is commonly used for coadds.
    survey: Survey to use for the zero-points and pixel scale.
    ra: Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: If `True`, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: Threshold used by `sep.extract()` to extract objects.
    use_mask: If `True`, the masked fits files are used. These must have
        been created beforehand.
    optimize_kronrad: If `True`, the Kron-radius scale is optimized, increasing the
        aperture size until the change in flux is less than ``eps``.
    eps: The Kron radius is increased until the change in flux is lower than ``eps``.
        A value of 0.0001 means 0.01% change in flux.
    gal_dist_thresh: Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    deblend_cont : Minimum contrast ratio used for object deblending. Default is 0.005.
        To entirely disable deblending, set to 1.0.
    save_plots: If `True`, the mask and galaxy aperture figures are saved.
    save_aperture_params: If `True`, the extracted mask parameters are saved into a pickle file.

    Returns
    -------
    gal_obj: Galaxy object.
    wcs: Image's WCS.
    kronrad: Kron radius.
    scale: Scale for the Kron radius.
    flip: Whether to flip the orientation of the aperture.
    """
    # initial checks
    check_survey_validity(survey)
    check_work_dir(workdir)
    obj_dir = Path(workdir, name)
    if use_mask:
        suffix = "_masked"
    else:
        suffix = ""
    if isinstance(filt, list):
        filt = "".join(f for f in filt)
    fits_file = obj_dir / survey / f"{survey}_{filt}{suffix}.fits"

    # image information
    hdu = fits.open(fits_file)
    hdu = remove_nan(hdu)
    header = hdu[0].header
    data = hdu[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        wcs = WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    # background subtraction, if needed
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg.back())
    else:
        data_sub = np.copy(data)

    # extract galaxy
    gal_obj, _ = extract_objects(
        data_sub,
        bkg.globalrms,
        host_ra,
        host_dec,
        threshold,
        wcs,
        gal_dist_thresh,
        deblend_cont,
    )
    if optimize_kronrad:
        exptime = get_image_exptime(header, survey)
        try:
            pixel_units = survey_pixel_units(survey, filt)
        except:
            # probably a coadd, so only need one filter
            pixel_units = survey_pixel_units(survey, filt[0])
        if pixel_units == "counts":
            _exptime = 1
        else:
            _exptime = exptime
        opt_res = optimize_aperture(data_sub, bkg.rms(), _exptime, gal_obj, eps)
        _, _, kronrad, scale = opt_res
    else:
        scale = 2.5
        kronrad, _ = sep.kron_radius(
            data_sub,
            gal_obj["x"],
            gal_obj["y"],
            gal_obj["a"],
            gal_obj["b"],
            gal_obj["theta"],
            6.0,
        )

    if save_plots is True:
        outfile = obj_dir / survey / f"global_{survey}_{filt}.jpg"
        title = rf"{name}: {survey}-${filt}$"
        plot_detected_objects(
            hdu,
            gal_obj,
            scale * kronrad,
            ra,
            dec,
            host_ra,
            host_dec,
            title,
            outfile,
        )

    if survey in flipped_surveys:
        flip = True
    else:
        flip = False

    if save_aperture_params is True:
        # save galaxy object and aperture parameters
        gal_df = pd.DataFrame(gal_obj)
        gal_df["kronrad"] = kronrad
        gal_df["scale"] = scale
        gal_df["flip"] = flip
        gal_df["filt"] = filt
        gal_df["survey"] = survey
        outfile = obj_dir / survey / f"aperture_parameters_{filt}.csv"
        gal_df.to_csv(outfile, index=False)
    hdu.close()

    return gal_obj, wcs, kronrad, scale, flip


def load_aperture_params(
    name: str, filt: str | list, survey: str
) -> tuple[np.ndarray, WCS, float, float, bool]:
    """Loads previously saved aperture parameters.

    Parameters
    ----------
    name: Name of the object to find the path of the aperture-parameters file.
    filt: Name of the filter used for the aperture parameters. Coadds are
        also valid.
    survey: Survey name to be used.

    Returns
    -------
    gal_obj, nongal_objs, wcs, sigma, r, flip: Mask parameters.
    """
    if isinstance(filt, list):
        filt = "".join(f for f in filt)
    obj_dir = Path(workdir, name)
    aper_params_file = obj_dir / survey / f"aperture_parameters_{filt}.csv"
    gal_df = pd.read_csv(aper_params_file)

    # split parameters
    kronrad = gal_df.pop("kronrad").values[0]
    scale = gal_df.pop("scale").values[0]  # remove host-galaxy row
    flip = gal_df.pop("flip").values[0]
    # remove unused columns
    _ = gal_df.pop("filt")
    _ = gal_df.pop("survey")
    # DataFrame to structured/record array
    gal_obj = gal_df.to_records()

    # load image WCS
    fits_file = obj_dir / survey / f"{survey}_{filt}.fits"
    hdu = fits.open(fits_file)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        wcs = WCS(hdu[0].header, naxis=2)

    return gal_obj, wcs, kronrad, scale, flip


def photometry(
    name: str,
    host_ra: float,
    host_dec: float,
    filt: str,
    survey: str,
    ra: Optional[float] = None,
    dec: Optional[float] = None,
    bkg_sub: Optional[bool] = None,
    threshold: float = 10,
    use_mask: bool = True,
    correct_extinction: float = True,
    optimize_kronrad: bool = True,
    eps: float = 0.0001,
    gal_dist_thresh: float = -1,
    deblend_cont: float = 0.005,
    common_aperture: bool = False,
    ref_filt: Optional[str] = None,
    ref_survey: Optional[str] = None,
    save_plots: bool = True,
    save_aperture_params: bool = False,
) -> tuple[float, float, float, float, float]:
    """Calculates the global aperture photometry of a galaxy using
    the Kron flux.

    **Note:** the galaxy must be ideally centred in the image.

    Parameters
    ----------
    name: Name of the object to find the path of the fits file.
    host_ra: Host-galaxy right ascension of the galaxy in degrees.
    host_dec: Host-galaxy declination of the galaxy in degrees.
    filt: Filter to use to load the fits file.
    survey: Survey to use for the zero-points and pixel scale.
    ra: Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: Threshold used by `sep.extract()` to extract objects.
    use_mask: If `True`, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    optimize_kronrad: If `True`, the Kron-radius scale is optimized, increasing the
        aperture size until the change in flux is less than ``eps``.
    eps: The Kron radius is increased until the change in flux is lower than ``eps``.
        A value of 0.0001 means 0.01% change in flux.
    gal_dist_thresh: Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    deblend_cont : Minimum contrast ratio used for object deblending. Default is 0.005.
        To entirely disable deblending, set to 1.0.
    common_aperture: Whether to use common aperture photometry using the reference filter(s) and survey.
    ref_filt: Reference filter (or coadd filters) from which to use the aperture parameters. Note that the parameters
        need to have been previously saved with ``save_aperture_params=True``.
    ref_survey: Reference survey from which to use the aperture parameters. Note that the parameters
        need to have been previously saved with ``save_aperture_params=True``.
    save_plots: If `True`, the mask and galaxy aperture figures are saved.
    save_aperture_params: Whether to save the aperture parameters.

    Returns
    -------
    mag: Aperture magnitude.
    mag_err: Error on the aperture magnitude.
    flux: Aperture flux.
    flux_err: Total flux error on the aperture flux.
    zp: Zeropoint.
    """
    # initial checks
    check_survey_validity(survey)
    check_work_dir(workdir)
    obj_dir = Path(workdir, name)
    if use_mask:
        suffix = "_masked"
    else:
        suffix = ""
    fits_file = obj_dir / survey / f"{survey}_{filt}{suffix}.fits"

    # image information
    hdu = fits.open(fits_file)
    hdu = remove_nan(hdu)
    header = hdu[0].header
    data = hdu[0].data
    exptime = get_image_exptime(header, survey)
    pixel_units = survey_pixel_units(survey, filt)
    if pixel_units == "counts":
        _exptime = 1
    else:
        _exptime = exptime
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        wcs = WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    # background subtraction, if needed
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg.back())
    else:
        data_sub = np.copy(data)

    if (ref_filt is None) | (ref_survey is None) | (common_aperture is False):
        # independent photometry for each filter
        gal_obj, wcs, kronrad, scale, _ = extract_aperture(
            name,
            host_ra,
            host_dec,
            filt,
            survey,
            ra,
            dec,
            bkg_sub,
            threshold,
            use_mask,
            optimize_kronrad,
            eps,
            gal_dist_thresh,
            deblend_cont,
            save_plots,
            save_aperture_params,
        )
    elif ((common_aperture is True) & (ref_filt is None)) | (
        (common_aperture is True) & (ref_survey is None)
    ):
        # error
        raise ValueError(
            "Reference filter(s) 'ref_filt' and survey 'ref_survey' must be given for common aperture photometry"
        )
    else:
        # common aperture photometry
        gal_obj, master_wcs, kronrad, scale, flip2 = load_aperture_params(
            name, ref_filt, ref_survey
        )
        if survey != ref_survey:
            # adapt different surveys
            if survey in flipped_surveys:
                flip = True
            else:
                flip = False
            if flip == flip2:
                flip_ = False
            else:
                flip_ = True
            # factor used for scaling the Kron radius between different pixel-scales / surveys
            gal_obj, conv_factor = adapt_aperture(
                gal_obj, master_wcs, wcs, flip_
            )
    # get Kron flux
    flux, flux_err = kron_flux(
        data_sub, bkg.rms(), _exptime, gal_obj, kronrad, scale
    )
    
    # aperture area for an ellipse (or circle)
    ap_area = np.pi * gal_obj["a"][0] * gal_obj["b"][0]
    mag, mag_err, flux, flux_err, zp = magnitude_calculation(
        flux,
        flux_err,
        survey,
        filt,
        ap_area,
        header,
        bkg.globalrms,
    )

    if correct_extinction is True:
        A_ext = calc_extinction(filt, survey, host_ra, host_dec)
        mag -= A_ext
        flux *= 10 ** (0.4 * A_ext)

    if save_plots is True:
        outfile = obj_dir / survey / f"global_{survey}_{filt}.jpg"
        filt_ = filt.replace("_", "-")
        title = f"{name}: {survey}-${filt_}$"
        plot_detected_objects(
            hdu,
            gal_obj,
            scale * kronrad,
            ra,
            dec,
            host_ra,
            host_dec,
            title,
            outfile,
        )
    hdu.close()

    return mag, mag_err, flux, flux_err, zp


def multi_band_phot(
    name: str,
    host_ra: float,
    host_dec: float,
    filters: Optional[str | list] = None,
    survey: str = "PanSTARRS",
    ra: Optional[float] = None,
    dec: Optional[float] = None,
    bkg_sub: Optional[bool] = None,
    threshold: float = 10,
    use_mask: bool = True,
    correct_extinction: bool = True,
    common_aperture: bool = False,
    ref_filt: Optional[str | list] = None,
    ref_survey: Optional[str | list] = None,
    optimize_kronrad: bool = True,
    eps: float = 0.0001,
    gal_dist_thresh: float = -1,
    deblend_cont: float = 0.005,
    save_plots: bool = True,
    save_aperture_params: bool = True,
    save_results: bool = True,
    raise_exception: bool = True,
    save_input: bool = True,
):
    """Calculates multi-band aperture photometry of the host galaxy
    for an object.

    Parameters
    ----------
    name: Name of the object to find the path of the fits file.
    host_ra: Host-galaxy right ascension of the galaxy in degrees.
    host_dec: Host-galaxy declination of the galaxy in degrees.
    filters: Filters to use to load the fits files. If `None` use all
        the filters of the given survey.
    survey: Survey to use for the zero-points and pixel scale.
    ra: Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: Threshold used by :func:`sep.extract()` to extract objects.
    use_mask: If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: If ``True``, the magnitudes are corrected for extinction.
    common_aperture: Whether to use common aperture photometry using the reference filter(s) and survey.
    ref_filt: Reference filter(s) used for common aperture photometry.
    ref_survey: Reference survey used for common aperture photometry. By default uses the same
        as the input 'survey'.
    optimize_kronrad: If `True`, the Kron-radius scale is optimized, increasing the
        aperture size until the change in flux is less than ``eps``.
    eps: The Kron radius is increased until the change in flux is lower than ``eps``.
        A value of 0.0001 means 0.01% change in flux.
        when optimizing the Kron radius.
    gal_dist_thresh: Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    deblend_cont : Minimum contrast ratio used for object deblending. Default is 0.005.
        To entirely disable deblending, set to 1.0.
    save_plots: If ``True``, the mask and galaxy aperture figures are saved.
    save_aperture_params: Whether to save the aperture parameters.
    save_results: If ``True``, the magnitudes are saved into a csv file.
    raise_exception: If ``True``, an exception is raised if the photometry fails for any filter.
    save_input: Whether to save the input parameters.

    Returns
    -------
    phot_df: DataFrame with the object's photometry and other info.

    Examples
    --------
    >>> from hostphot.photometry import global_photometry as gp
    >>> name = 'SN2004eo'
    >>> host_ra, host_dec = 308.2092, 9.92755  # coords of host galaxy of SN2004eo
    >>> ra, dec =  308.22579, 9.92853 # coords of SN2004eo
    >>> results = gp.multi_band_phot(name, 
                                     host_ra, 
                                     host_dec, 
                                     survey=survey, 
                                     ra=ra, 
                                     dec=dec, 
                                     use_mask=True,
                                     common_aperture=True, 
                                     ref_filt='riz',
                                     ref_survey='PanSTARRS', 
                                     save_plots=True)
    """
    input_params = locals()  # dictionary
    check_survey_validity(survey)
    if filters is None:
        if survey in ["HST", "JWST"]:
            raise ValueError(f"For {survey}, the filter needs to be specified!")
        filters = get_survey_filters(survey)
    else:
        check_filters_validity(filters, survey)
    if survey in ["HST", "JWST"]:
        filters = [filters]

    # save input parameters
    if save_input is True:
        inputs_file = Path(workdir, name, survey, "input_global_photometry.csv")
        store_input(input_params, inputs_file)

    results_dict = {
        "name": name,
        "host_ra": host_ra,
        "host_dec": host_dec,
        "survey": survey,
    }

    if ref_survey is None:
        ref_survey = survey
    """
    if (common_aperture is True) & (ref_filt is not None):
        # use common aperture
        _ = extract_aperture(
            name,
            host_ra,
            host_dec,
            ref_filt,
            ref_survey,
            ra=ra,
            dec=dec,
            bkg_sub=bkg_sub,
            threshold=threshold,
            use_mask=use_mask,
            optimize_kronrad=optimize_kronrad,
            eps=eps,
            gal_dist_thresh=gal_dist_thresh,
            deblend_cont=deblend_cont,
            save_plots=save_plots,
            save_aperture_params=True,
        )
    elif (common_aperture is True) & (ref_filt is None):
        raise ValueError(
            "Reference filter(s) 'ref_filt' must be given for common aperture"
        )
    """

    for filt in filters:
        try:
            mag, mag_err, flux, flux_err, zp = photometry(
                name,
                host_ra,
                host_dec,
                filt,
                survey,
                ra=ra,
                dec=dec,
                bkg_sub=bkg_sub,
                threshold=threshold,
                use_mask=use_mask,
                correct_extinction=correct_extinction,
                optimize_kronrad=optimize_kronrad,
                eps=eps,
                gal_dist_thresh=gal_dist_thresh,
                deblend_cont=deblend_cont,
                common_aperture=common_aperture,
                ref_filt=ref_filt,
                ref_survey=ref_survey,
                save_plots=save_plots,
                save_aperture_params=save_aperture_params,
            )
            results_dict[filt] = mag
            results_dict[f"{filt}_err"] = mag_err
            results_dict[f"{filt}_flux"] = flux
            results_dict[f"{filt}_flux_err"] = flux_err
            results_dict[f"{filt}_zeropoint"] = zp
        except Exception as exc:
            if raise_exception is True:
                raise Exception(f"{filt}-band: {exc}")
            else:
                results_dict[filt] = np.nan
                results_dict[f"{filt}_err"] = np.nan
                results_dict[f"{filt}_flux"] = np.nan
                results_dict[f"{filt}_flux_err"] = np.nan
                results_dict[f"{filt}_zeropoint"] = np.nan

    phot_df = pd.DataFrame({key: [val] for key, val in results_dict.items()})
    if save_results is True:
        outfile = Path(workdir, name, survey, "global_photometry.csv")
        phot_df.to_csv(outfile, index=False)

    return phot_df



from astropy import wcs
import astropy.units as u
from astropy.io import fits
from astropy.coordinates import SkyCoord
from astroquery.ukidss import Ukidss

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

import warnings
from astropy.utils.exceptions import AstropyWarning

def get_UKIDSS_images(ra: float, dec: float, size: float=3, filters: str='ZYJHK') -> list[fits.ImageHDU]:
    """Gets UKIDSS fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``ZYJHK``.

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "UKIDSS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    
    database = 'UKIDSSDR11PLUS'
    # programme = 'LAS'  # ['UDS', 'GCS', 'GPS', 'DXS', 'LAS']
    
    survey_pixel_scale(survey, "H")  # same scale for all pixels, except J for LAS
    if isinstance(size, (float, int)):
        size = (size * u.arcmin)

    u_obj = Ukidss(database=database)
    pos = SkyCoord(ra, dec, unit="deg")
    urls = u_obj.get_image_list(pos, waveband='all', frame_type="stack", image_width=size)

    hdu_dict = {filt:None for filt in filters}
    for filt in filters:
        for url in urls:
            # pick the first FITS image only (is this correct?)
            if f'band={filt}' in url:
                hdu = fits.open(url)
                # update first extension with data and WCS
                hdu[0].data = hdu[1].data
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", AstropyWarning)
                    img_wcs = wcs.WCS(hdu[1].header)
                    hdu[0].header.update(img_wcs.to_header())
                # add some of the keywords to the PHU
                hdu[0].header['MAGZRR'] = hdu[1].header['MAGZRR']
                hdu[0].header['GAIN'] = hdu[1].header['GAIN']
                hdu[0].header['READNOIS'] = hdu[1].header['READNOIS']
                # calculate effective ZP (considering atmospheric extinction)
                # calculate extinction first
                airmass = (hdu[0].header['AMSTART'] + hdu[0].header['AMEND'])/2
                ext_coeff = hdu[1].header['EXTINCT']
                extinction = ext_coeff*(airmass - 1)
                # calculate effective ZP
                zp = hdu[1].header['MAGZPT']
                hdu[0].header['MAGZP'] = zp - extinction
                hdu_dict[filt] = hdu
                break                
    hdu_list = list(hdu_dict.values()) 
    return hdu_list


import numpy as np
from pathlib import Path

import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astropy.coordinates import SkyCoord

from astroquery.esa.jwst import Jwst

from hostphot._constants import workdir
from hostphot.utils import check_work_dir, suppress_stdout
from hostphot.surveys_utils import check_JWST_filters, survey_pixel_scale

import warnings
from astropy.utils.exceptions import AstropyWarning


def update_JWST_header(hdu: fits.hdu.ImageHDU) -> None:
    """Updates the JWST image header with the necessary keywords.

    Parameters
    ----------
    hdu : JWST FITS image.
    """
    # get WCS - SCI is hdu[1] - hostphot always assumes hdu[0] is used, so need to move them over
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = WCS(hdu[1].header)
    hdu[0].header.update(img_wcs.to_header())
    hdu[0].header["PIXAR_SR"] = hdu[1].header["PIXAR_SR"]
    hdu[0].data = hdu[1].data
    # add zeropoints
    # https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-performance/nircam-absolute-flux-calibration-and-zeropoints
    pixar_sr = hdu[0].header["PIXAR_SR"]
    hdu[0].header["MAGZP"] = (
       -6.10 - (2.5 * np.log10(pixar_sr))
    )

def set_JWST_image(file: str, filt: str, name: str) -> None:
    """Moves a previously downloaded JWST image into the work directory.

    The image's header is updated with the necessary keywords to obtain
    photometry and is also moved under the objects directory inside the
    work directory.

    JWST images take very long to download, so the user might prefer to
    download the images manually and then use this function to include
    the image into the workflow.

    Parameters
    ----------
    file: JWST image to use.
    filt: JWST filter, e.g. ``NIRCam_F150W``.
    name: Object's name.
    """
    # check output directory
    check_work_dir(workdir)
    check_JWST_filters(filt)
    obj_dir = Path(workdir, name, "JWST")
    if obj_dir.is_dir() is False:
        obj_dir.mkdir(parents=True, exist_ok=True)
    # update header and save file
    hdu = fits.open(file)
    update_JWST_header(hdu)
    outfile = obj_dir / f"JWST_{filt}.fits"
    hdu.writeto(outfile, overwrite=True)
    
def get_JWST_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                        filters: list = ["WFC3_UVIS_F225W"]) -> list[fits.ImageHDU]:
    """Downloads a set of JWST fits images for a given set
    of coordinates and filters using astroquery.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use, e.g. ``WFC3_UVIS_F225W``.

    Return
    ------
    hdu_list: List with fits image for the given filter. ``None`` is returned if no image is found.
    """
    Jwst.get_status_messages()
    
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)
    size_arcsec = size_arcsec.value
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

    # query observations at the given coordinates
    with suppress_stdout():
        result = Jwst.cone_search(
            radius=3*u.arcsec,
            coordinate=coords,
            cal_level="Top",
            prod_type="image",
            #instrument_name=instrument,
            #filter_name=filt,
            only_public=True,
            async_job=True,
        ).get_data()
    results_df = result.to_pandas()
            
    hdu_list = []
    for filt in filters:
        check_JWST_filters(filt)
        # separate the instrument name from the actual filter
        split_filt = filt.split("_")
        filt = split_filt[-1]
        instrument = split_filt[0].upper() + "/IMAGE"
        # filter by filter and instrument
        obs_df = results_df[results_df["instrument_name"] == instrument]
        obs_df = obs_df[obs_df["energy_bandpassname"] == filt]
        obs_df = obs_df[obs_df["calibrationlevel"] == 3]

        # start download
        temp_file = None
        for obs_id in obs_df.observationid:
            try:
                product_list = Jwst.get_product_list(observation_id=obs_id, product_type='science', cal_level=3).to_pandas()
                # choose first file as there es no exptime info and the images take too lonk to download multiple ones
                file_name = [file for file in product_list.filename if file.endswith(".fits")][0] 
                temp_file = Jwst.get_product(file_name=file_name)
                break
            except:
                pass
        if temp_file is None:
            hdu_list.append(None)
            continue
        hdu = fits.open(temp_file)
        # remove the temporary files
        Path(temp_file).unlink()
        # add necessary information to the header
        update_JWST_header(hdu)
        hdu_list.append(hdu)
    # JWST images are large so need to be trimmed
    for hdu, filt in zip(hdu_list, filters):
        if hdu is None:
            continue
        pixel_scale = survey_pixel_scale("JWST", filt)  # same pixel scale for all filters
        size_pixels = int(size_arcsec / pixel_scale)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            img_wcs = WCS(hdu[0].header)
        trimmed_data = Cutout2D(hdu[0].data, coords, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())
    return hdu_list


import numpy as np
from typing import Optional

from astropy.wcs import WCS
import astropy.units as u
from astropy.io import fits
from astropy.nddata import Cutout2D
from astropy.coordinates import SkyCoord
from astroquery.mast import Observations

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

import warnings
from astropy.utils.exceptions import AstropyWarning

def get_GALEX_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: Optional[str] = None, version: Optional[str] = None) -> list[fits.ImageHDU]:
    """Downloads a set of GALEX fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``FUV, NUV``.
    version: Version of GALEX images. Either Deep (``DIS``), Medium (``MIS``) or
             All-Sky Imaging Survey (``AIS``), or Nearby Galaxy Survey (``NGS``) or
             Guest Investigator Survey (``GII``). If ``None``, take the image with the
             longest exposure time.

    Returns
    -------
    hdu_list: List with fits images for the given filters. ``None`` is returned if no image is found.
    """
    survey = "GALEX"
    if filters is None:
        filters = get_survey_filters(survey)
    if isinstance(filters, str):
        filters = [filters]
    check_filters_validity(filters, survey)
    # get size in pixels
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
    obs_table = Observations.query_criteria(
        coordinates=coords, radius=size_arcsec, obs_collection=["GALEX"]
    )
    obs_df_ = obs_table.to_pandas()

    if version is None:
        # starts from the survey with the deepest images first
        obs_df_ = obs_df_.sort_values("t_exptime", ascending=False)
        projects = obs_df_.project.unique()
    else:
        # only use the survey requested by the user
        projects = [version]

    hdu_dict = {"NUV": None, "FUV": None}
    for project in projects:
        obs_df = obs_df_[obs_df_.project == project]

        for filt in filters:
            if hdu_dict[filt] is not None:
                # if the image was already found, skip this filter
                continue

            # get only "intensity" images
            filt_extension = {
                "NUV": "-nd-int.fits.gz",
                "FUV": "-fd-int.fits.gz",
            }
            # get unique image sectors
            files = []
            for file in obs_df.dataURL.values:
                sector_info = file.split("-")[:-2]
                file = (
                    "-".join(string for string in sector_info)
                    + filt_extension[filt]
                )
                if file not in files:
                    files.append(file)

            # download the FITS images
            hdu_list = []
            for file in files:
                try:
                    hdu = fits.open(file)
                    hdu_list.append(hdu)
                except:
                    pass

            # calculate the separation of the galaxy to the image center
            separations = []
            for hdu in hdu_list:
                ra_img = float(hdu[0].header["RA_CENT"])
                dec_img = float(hdu[0].header["DEC_CENT"])
                coords_img = SkyCoord(
                    ra=ra_img,
                    dec=dec_img,
                    unit=(u.degree, u.degree),
                    frame="icrs",
                )
                separation = coords.separation(coords_img).value
                separations.append(separation)

            # get the image with the galaxy closest to the center
            if len(separations) != 0:
                id_file = np.argmin(separations)
                hdu = hdu_list[id_file]
            else:
                hdu = None

            hdu_dict[filt] = hdu

    hdu_list = []
    for filt in filters:
        pixel_scale = survey_pixel_scale(survey, filt)
        size_pixels = int(size_arcsec.value / pixel_scale)
    
        hdu = hdu_dict[filt]
        if hdu is None:
            hdu_list.append(None)
            continue  # no image in this filter
        # trim data to requested size
        img_wcs = WCS(hdu[0].header)
        pos = SkyCoord(ra=ra * u.degree, dec=dec * u.degree)

        trimmed_data = Cutout2D(hdu[0].data, pos, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())
        hdu_list.append(hdu)
    return hdu_list


import unittest
from pathlib import Path
from hostphot.cutouts import download_images, set_HST_image, set_JWST_image


class TestHostPhot(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestHostPhot, self).__init__(*args, **kwargs)
        # object used for most surveys
        self.sn_name = "2002fk"
        self.ra = 50.527333
        self.dec = -15.400056

    def test_cutouts_PanSTARRS(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="PanSTARRS"
        )

    def test_cutouts_DES(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="DES"
        )

    def test_cutouts_SDSS(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="SDSS"
        )

    def test_cutouts_GALEX(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="GALEX"
        )

    def test_cutouts_WISE(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="WISE"
        )

    def test_cutouts_2MASS(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="2MASS"
        )

    def test_cutouts_unWISE(self):
        for version in ["neo1", "neo2", "allwise"]:
            download_images(
                self.sn_name,
                self.ra,
                self.dec,
                overwrite=True,
                survey="unWISE",
                version=version,
            )

    def test_cutouts_LegacySurvey(self):
        download_images(
            self.sn_name,
            self.ra,
            self.dec,
            overwrite=True,
            survey="LegacySurvey",
        )

    def test_cutouts_Spitzer(self):
        name = "Spitzer_test"
        ra, dec = 52.158591, -27.891113
        download_images(name, ra, dec, overwrite=True, survey="Spitzer")

    def test_cutouts_VISTA(self):
        name = "VISTA_test"
        # use different coordinates for each survey as they don't overlap
        surveys = {
            "VHS": [120, -60],
            "VIDEO": [36.1, -5],
            "VIKING": [220.5, 0.0],
        }

        
        for version, coords in surveys.items():
            ra, dec = coords
            #try:
            download_images(
                name,
                ra,
                dec,
                overwrite=True,
                survey="VISTA",
                version=version,
            )

    def test_cutouts_SkyMapper(self):
        download_images(
            self.sn_name,
            self.ra,
            self.dec,
            overwrite=True,
            survey="SkyMapper",
        )

    def test_cutouts_SPLUS(self):
        name = "SPLUS_test"
        ra, dec = 0.6564206, -0.3740297
        download_images(name, ra, dec, overwrite=True, survey="SPLUS")

    def test_cutouts_UKIDSS(self):
        name = "UKIDSS_test"
        ra, dec = 359.5918320, +0.1964120
        download_images(name, ra, dec, overwrite=True, survey="UKIDSS")
        
    def test_cutouts_HST(self):
        name = "HST_test"
        filt = 'WFC3_UVIS_F275W'
        file = "tests/hst_16741_2v_wfc3_uvis_f275w_iepo2v_drc.fits"
        if Path(file).exists():
            # this test only runs locally as the file is too large
            set_HST_image(file, filt, name)
        
    def test_cutouts_JWST(self):
        name = "JWST_test"
        filt = 'NIRCam_F090W'
        file = "tests/jw01685-c1008_t004_nircam_clear-f090w_i2d.fits"
        if Path(file).exists():
            # this test only runs locally as the file is too large
            set_JWST_image(file, filt, name)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from typing import Optional

from astropy.wcs import WCS
from astropy.io import fits
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import Cutout2D
from astroquery.sdss import SDSS

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

import warnings
from astropy.utils.exceptions import AstropyWarning

def get_SDSS_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: Optional[str] = None, version: Optional[str] = None) -> fits.HDUList | None:
    """Downloads a set of SDSS fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``ugriz``.
    version: Data release version to use. If not given, use the latest one (``dr17``).

    Returns
    -------
    hdu_list: List with fits images for the given filters. ``None`` is returned if no image is found.
    """
    # check filters
    survey = "SDSS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    # check data release version
    if version is None:
        version = "dr17"
    versions = [f"dr{i}" for i in range(12, 17 + 1)]
    assert (
        version in versions
    ), f"The given version ({version}) is not a valid data release: {versions}"
    dr = int(version.replace("dr", ""))
    # get size in pixels
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)
    pixel_scale = survey_pixel_scale(survey, "g")  # same pixel scale for all filters
    size_pixels = int(size_arcsec.value / pixel_scale)
    # get SDSS ids near the given coordinates
    coords = SkyCoord(ra=ra * u.degree, dec=dec * u.degree)
    for radius in np.arange(1, 60, 1):
        ids = SDSS.query_region(
            coords, radius=radius * u.arcsec, data_release=dr
        )
        if ids is not None:
            if 'ra' in ids.colnames and 'dec' in ids.colnames:
                break
            else:
                ids = None
    if ids is None:
        return None
    # get the pointing closest to the given coordinates
    coords_imgs = SkyCoord(
        ra=ids["ra"].value,
        dec=ids["dec"].value,
        unit=(u.degree, u.degree),
        frame="icrs",
    )
    separation = coords.separation(coords_imgs).value
    pointing_id = np.argmin(separation)
    ids2remove = list(np.arange(len(separation)))
    del ids2remove[pointing_id]
    ids.remove_rows(ids2remove)
    # download images
    hdu_list = SDSS.get_images(matches=ids, band=filters, data_release=dr)
    # SDSS images are large so need to be trimmed
    for hdu in hdu_list:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            img_wcs = WCS(hdu[0].header)
        trimmed_data = Cutout2D(hdu[0].data, coords, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())
    return hdu_list


import re
import urllib
from typing import Optional

from astropy import wcs
import astropy.units as u
from astropy.io import fits

from hostphot.surveys_utils import get_survey_filters, check_filters_validity
from hostphot.moc.maps import contains_coords

import warnings
from astropy.utils.exceptions import AstropyWarning


def get_VISTA_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                        filters: Optional[str] = None) -> list[fits.ImageHDU]:
    """Gets VISTA fits images for the given coordinates and
    filters.

    Note: the different surveys included in VISTA cover different
    parts of the sky and do not necessarily contain the same filters.
    
    The available surveys are "VIDEO", "VIKING", "VHS", and chosen in
    that order, depending on the coverage.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``Z, Y, J, H, Ks``.
    version: str, default ``VHS``
        Survey to use: ``VHS``, ``VIDEO`` or ``VIKING``.

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "VISTA"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    if not isinstance(size, (float, int)):
        size = size.to(u.arcmin).value

    for version in ["VIDEO", "VIKING", "VHS"]:
        overlap = contains_coords(ra, dec, version)
        if overlap is True:
            break
    if overlap is False:
        return [False] * len(filters)
    
    # Latest data releases
    database_dict = {
        "VHS": "VHSDR6",
        "VIDEO": "VIDEODR6",
        "VIKING": "VIKINGDR5",
    }
    database = database_dict[version]

    base_url = "http://vsa.roe.ac.uk:8080/vdfs/GetImage?archive=VSA&"
    survey_dict = {
        "database": database,
        "ra": ra,
        "dec": dec,
        "sys": "J",
        "filterID": "all",
        "xsize": size,  # in arcmin
        "ysize": size,  # in arcmin
        "obsType": "object",
        "frameType": "tilestack",
    }
    survey_url = "&".join([f"{key}={val}" for key, val in survey_dict.items()])
    url = base_url + survey_url
    results = urllib.request.urlopen(url).read()
    links = re.findall('href="(http://.*?)"', results.decode("utf-8"))
    
    # find url for each filter (None if not found)
    urls_dict = {filt: [] for filt in filters}
    for filt in filters:
        for link in links:
            url = link.replace("getImage", "getFImage", 1)
            if f"band={filt}" in url:
                urls_dict[filt].append(url)
    
    hdu_list = []
    for filt, url_list in urls_dict.items():
        if len(url_list) > 0:
            exptime = 0
            # select image with longest exposure
            for url in url_list:
                hdu_ = fits.open(url)
                if hdu_[1].header["EXPTIME"] > exptime:
                    exptime = hdu_[1].header["EXPTIME"]
                    hdu = hdu_
                else:
                    hdu_.close()
            hdu[0].data = hdu[1].data
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", AstropyWarning)
                img_wcs = wcs.WCS(hdu[1].header)
                hdu[0].header.update(img_wcs.to_header())
            # add some keywords to the PHU
            hdu[0].header['EXPTIME'] = hdu[1].header['EXPTIME']
            hdu[0].header['MAGZRR'] = hdu[1].header['MAGZRR']
            # calculate effective ZP (considering atmospheric extinction)
            # calculate extinction first
            airmass = (hdu[1].header['HIERARCH ESO TEL AIRM START'] + hdu[1].header['HIERARCH ESO TEL AIRM END'])/2
            ext_coeff = hdu[1].header['EXTINCT']
            extinction = ext_coeff*(airmass - 1)
            # calculate effective ZP
            zp = hdu[1].header['MAGZPT']
            hdu[0].header['MAGZP'] = zp - extinction
            hdu_list.append(hdu)
        else:
            hdu_list.append(None)
    return hdu_list


import astropy.units as u
from astropy.io import fits
from typing import Optional

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

def get_LegacySurvey_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                        filters: Optional[str] = None, version: str = "dr10") -> list[list[fits.HDUList]]:
    """Gets Legacy Survey fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``grz``.
    version: Data release version. E.g. ``dr10``, ``dr9``, ``dr8``, etc..

    Returns
    -------
    fits_files: List of fits images.
    """
    survey = "LegacySurvey"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    if isinstance(filters, list):
        filters = "".join(filter for filter in filters)

    pixel_scale = survey_pixel_scale(survey, "g")  # same pixel scale for all filters
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    if version is None:
        version = "dr10"  # latest data release

    base_url = "https://www.legacysurvey.org/viewer/fits-cutout?"
    params = f"ra={ra}&dec={dec}&layer=ls-{version}&pixscale={pixel_scale}&bands={filters}&size={size_pixels}&invvar"
    url = base_url + params

    master_hdu = fits.open(url)
    master_header = master_hdu[0].header
    master_header_invvar = master_hdu[1].header

    hdu_list = []
    for i, filt in enumerate(filters):
        data = master_hdu[0].data[i]
        header = master_header.copy()
        header.append(('BAND', filt, ' Band - added by HostPhot'), end=True)
        hdu = fits.PrimaryHDU(data=data, header=header)
        
        header_invvar = master_header_invvar.copy()
        header_invvar.append(('BAND', filt, ' Band - added by HostPhot'), end=True)
        data_invvar = master_hdu[1].data[i]
        hdu_invvar = fits.ImageHDU(data=data_invvar, 
                                header=header_invvar)
        hdu_list.append(fits.HDUList([hdu, hdu_invvar]))
    master_hdu.close()
    return hdu_list


from mocpy import MOC
import astropy.units as u
from astropy.coordinates import SkyCoord
import warnings

import hostphot
from pathlib import Path
mapsdir = Path(hostphot.__path__[0], 'moc/maps')

    
def contains_coords(ra: float, dec: float, survey: str) -> bool:
    """Checks whether the given coordinates overlap with a survey footprint.

    Parameters
    ----------
    ra: right ascension in degrees.
    dec: declination in degrees.
    survey: survey name.

    Returns
    -------
    overlap: whether the coordinates overlap.
    """
    # load MOC map
    moc_file = [file for file in mapsdir.glob(f"{survey}*.fits")][0]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        moc_map = MOC.load(moc_file)
    # check intersection with given coordinates
    coord = SkyCoord(ra * u.deg, dec * u.deg)
    overlap = moc_map.contains_skycoords(coord)
    if overlap:
        return True
    else:
        return False



import re
import requests
from typing import Optional

import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astropy.coordinates import SkyCoord

from hostphot.surveys_utils import get_survey_filters, check_filters_validity, survey_pixel_scale

    
def get_2MASS_images(
    ra: float, dec: float, size: float | u.Quantity = 3, filters: Optional[str] = None
) -> fits.HDUList:
    """Downloads a set of 2MASS fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters:  Filters to use. If ``None``, uses ``FUV, NUV``.

    Return
    ------
    hdu_list: List with fits images for the given filters. ``None`` is returned if no image is found.
    """
    survey = "2MASS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_degree = (size * u.arcmin).to(u.degree)
    else:
        size_degree = size.to(u.degree)
    size_degree = size_degree.value
    coords = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    
    # this code was taken from chatgpt and combined with that from Blast
    # Base URL for 2MASS cutout service
    base_url = "https://irsa.ipac.caltech.edu/cgi-bin/2MASS/IM/nph-im_sia"
    # Query parameters
    params = {
        "POS": f"{ra},{dec}",
        #"SIZE": size_degree, 
        "SIZE": 0.01, 
        "FORMAT": "image/fits"
    }
    # Make the request to the 2MASS server
    response = requests.get(base_url, params=params)
    response.raise_for_status()
    
    hdu_list = []
    for filter in filters:
        if filter == "Ks":
            filter = "K"
        fits_image = None
        for line in response.content.decode("utf-8").split("<TD><![CDATA["):
            if re.match(f"https://irsa.*{filter.lower()}i.*fits", line.split("]]>")[0]):
                fitsurl = line.split("]]")[0]
                fits_image = fits.open(fitsurl, cache=None)
                wcs = WCS(fits_image[0].header)
                if coords.contained_by(wcs):
                    break
        else:
            fits_image = None
        hdu_list.append(fits_image)

    for filt, hdu in zip(filters, hdu_list):
        pixel_scale = survey_pixel_scale(survey, filt)
        size_pixels = int((size_degree * u.deg).to(u.arcsec).value / pixel_scale)
    
        wcs = WCS(hdu[0].header)
        cutout = Cutout2D(hdu[0].data, coords, size_pixels, wcs=wcs)
        hdu[0].data = cutout.data
        hdu[0].header.update(cutout.wcs.to_header())
        
    return hdu_list



# -*- coding: utf-8 -*-

from ._version import __version__
from ._constants import workdir
from .utils import plot_fits, plot_image
from .cutouts.downloads import download_images
from .photometry import local_photometry
from .photometry import global_photometry


import numpy as np
import pandas as pd

import astropy.units as u
from astropy.io import fits

from pyvo.dal import sia
from hostphot.surveys_utils import get_survey_filters, check_filters_validity

def get_SkyMapper_urls(ra: float, dec: float, fov: float | u.Quantity = 3, 
                    filters: str = "uvgriz") -> list[str] | None:
    """Obtains the URLs of the SkyMapper images.
    
    The images closest to the given coordinates are retrieved 

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    fov: Field of view in degrees.
    filters: Filters to use. If ``None``, uses ``uvgriz``.

    Returns
    -------
    url_list: List of URLs with SkyMapper images.
    """
    sm_url = "https://api.skymapper.nci.org.au/public/siap/dr4/query"
    svc = sia.SIAService(sm_url)
    imgs_table = svc.search(
        (ra, dec), (fov / np.cos(dec * np.pi / 180), fov), verbosity=2
    )
    if len(imgs_table) == 0:
        print(("Warning: empty table returned for " f"ra={ra}, dec={dec}"))
        return None
    imgs_df = pd.DataFrame(imgs_table)
    imgs_df = imgs_df[imgs_df.format.str.endswith('fits')]
    if len(imgs_df) == 0:
        return None

    url_list = []
    for filt in filters:
        filt_df = imgs_df[imgs_df.band==filt]
        # obtain the images with largest exposure times and file size
        # and lowest mean FWHM, as suggested by Christopher Onken, from the SkyMapper Team
        filt_df = filt_df[filt_df.exptime==filt_df.exptime.max()]
        filt_df = filt_df[filt_df.filesize==filt_df.filesize.max()]
        filt_df = filt_df[filt_df.mean_fwhm==filt_df.mean_fwhm.min()]
        if len(filt_df) == 0:
            url_list.append(None)
        else:
            fits_url = filt_df.get_fits.values[0]
            url_list.append(fits_url)    
    return url_list

def get_SkyMapper_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: str = "uvgriz") -> fits.HDUList | None:
    """Gets SkyMapper DR4 fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses ``uvgriz``.

    Returns
    -------
    hdu_list: List of fits images.
    """
    # initial checks
    survey = "SkyMapper"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    if isinstance(size, (float, int)):
        fov = (size * u.arcmin).to(u.degree).value
    else:
        fov = size.to(u.degree).value
    # download images
    url_list = get_SkyMapper_urls(ra, dec, fov, filters)
    if url_list is None:
        return None
    hdu_list = []
    for url in url_list:
        if url is None:
            hdu_list.append(None)
        else:
            hdu = fits.open(url)
            # add zeropoint
            hdu[0].header['MAGZP'] = hdu[0].header['ZPAPPROX']
            hdu_list.append(hdu)
    return hdu_list


import numpy as np
import pandas as pd
from pathlib import Path
from typing import Optional

from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from photutils.utils import calc_total_error

import hostphot
from hostphot.surveys_utils import (
    check_survey_validity,
    check_filters_validity,
    survey_pixel_units,
    survey_zp,
)
from hostphot.photometry.image_utils import (
    get_image_exptime,
    get_image_gain,
    get_image_readnoise,
    correct_HST_aperture,
)

hostphot_path = Path(hostphot.__path__[0])


def calc_sky_unc(image: np.ndarray, exptime: float) -> float:
    """Calculates the uncertainty of the image from the
    sky's standard deviation, sigma-clipped STD.

    Parameters
    ----------
    image: Image in a 2D numpy array.
    exptime: Exposure time of the image.

    Returns
    -------
    error: Estimated error of the image.
    """
    mask = image != 0
    avg, sky, sky_std = sigma_clipped_stats(image[mask], sigma=3.0)
    error = calc_total_error(image, sky_std, exptime)

    return error


def get_HST_err(filt: str, header: fits.Header) -> float:
    """Obtaines the error propagation from the zeropoint.

    Parameters
    ----------
    filt: HST filter, e.g. ``WFC3_UVIS_F225W``.
    header: Header of an image.

    Returns
    -------
    mag_err: Magnitude error on PHOTFLAM.
    """
    # split instrument and filter
    filt_split = filt.split("_")
    filt = filt_split[-1]
    instrument = filt_split[-2]

    if instrument == "UVIS":
        # APERTURE usually points to UVIS2
        instrument = header["APERTURE"]
        # some images have a different APERTURE value
        # see: https://hst-docs.stsci.edu/wfc3ihb/chapter-6-uvis-imaging-with-wfc3/6-4-uvis-field-geometry
        # not sure if this is the correct solution
        if instrument == "UVIS-CENTER":
            instrument = "UVIS2"
        if instrument == "UVIS":
            instrument = "UVIS1"

    # get uncertainty file
    err_file = hostphot_path.joinpath("filters", "HST", f"{instrument}_err.txt")
    err_df = pd.read_csv(err_file, sep="\\s+")
    filt_err_df = err_df[err_df.Filter == filt]
    # error propagation
    flux = filt_err_df.PHOTFLAM.values[0]
    flux_err = filt_err_df.ERR_PHOTFLAM.values[0]
    mag_err = np.abs(2.5 * flux_err / (flux * np.log(10)))

    return mag_err


def uncertainty_calculation(
    flux: float,
    flux_err: float,
    survey: str,
    filt: Optional[str] = None,
    ap_area: float = 0.0,
    header: Optional[fits.Header] = None,
    bkg_rms: float = 0.0,
) -> float:
    """Calculates the uncertainty propagation.

    Parameters
    ----------
    flux: Aperture flux.
    flux_err: Aperture flux error.
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.
    filt: Survey-specific filter.
    ap_area: Aperture area.
    header: Header of an image.
    bkg_rms: Background noise.

    Returns
    -------
    flux_err: Total uncertainty in flux units.
    """
    exptime = get_image_exptime(header, survey)
    gain = get_image_gain(header, survey)
    readnoise = get_image_readnoise(header, survey)

    # 1.0857 = 2.5/ln(10)
    mag_err = 2.5 / np.log(10) * flux_err / flux

    if survey in [
        "PanSTARRS",
        "DES",
        "LegacySurvey",
        "Spitzer",
        "VISTA",
        "SkyMapper",
        "SPLUS",
        "UKIDSS",
    ]:
        if survey == "Spitzer":
            flux /= header["EFCONV"]  # conv. factor (MJy/sr)/(DN/s)

        extra_err = (
            2.5 / np.log(10) * np.sqrt(ap_area * (readnoise**2) + flux / gain) / flux
        )
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

    if survey == "DES":
        # see the photometry section in https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/quality
        # statistical uncertainties on the AB magnitud system zeropoints
        unc_dict = {
            "g": 2.6e-3,
            "r": 2.9e-3,
            "i": 3.4e-3,
            "z": 2.5e-3,
            "Y": 4.5e-3,
        }
        extra_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

        # median coadd zeropoint statistical uncertainty
        unc_dict = {
            "g": 5e-3,
            "r": 4e-3,
            "i": 5e-3,
            "z": 6e-3,
            "Y": 5e-3,
        }
        extra_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

    elif survey == "PanSTARRS":
        # add floor systematic error from:
        # https://iopscience.iop.org/article/10.3847/1538-4365/abb82a/pdf
        unc_dict = {
            "g": 14e-3,
            "r": 14e-3,
            "i": 15e-3,
            "z": 15e-3,
            "y": 18e-3,
        }
        floor_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err**2 + floor_err**2)

    elif survey == "SDSS":
        # https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
        camcol = header["CAMCOL"]
        run = header["RUN"]

        gain_dict = {
            "u": {1: 1.62, 2: 1.595, 3: 1.59, 4: 1.6, 5: 1.47, 6: 2.17},
            "g": {1: 3.32, 2: 3.855, 3: 3.845, 4: 3.995, 5: 4.05, 6: 4.035},
            "r": {1: 4.71, 2: 4.6, 3: 4.72, 4: 4.76, 5: 4.725, 6: 4.895},
            "i": {1: 5.165, 2: 6.565, 3: 4.86, 4: 4.885, 5: 4.64, 6: 4.76},
            "z": {1: 4.745, 2: 5.155, 3: 4.885, 4: 4.775, 5: 3.48, 6: 4.69},
        }
        # dark variance
        dv_dict = {
            "u": {
                1: 9.61,
                2: 12.6025,
                3: 8.7025,
                4: 12.6025,
                5: 9.3025,
                6: 7.0225,
            },
            "g": {
                1: 15.6025,
                2: 1.44,
                3: 1.3225,
                4: 1.96,
                5: 1.1025,
                6: 1.8225,
            },
            "r": {
                1: 1.8225,
                2: 1.00,
                3: 1.3225,
                4: 1.3225,
                5: 0.81,
                6: 0.9025,
            },
            "i": {1: 7.84, 2: 5.76, 3: 4.6225, 4: 6.25, 5: 7.84, 6: 5.0625},
            "z": {1: 0.81, 2: 1.0, 3: 1.0, 4: 9.61, 5: 1.8225, 6: 1.21},
        }
        gain = gain_dict[filt][camcol]
        dark_variance = dv_dict[filt][camcol]

        if filt == "u" and camcol == 2 and run > 1100:
            gain = 1.825
        if filt == "i" and run > 1500:
            if camcol == 2:
                dark_variance = 6.25
            if camcol == 4:
                dark_variance = 7.5625
        if filt == "z" and run > 1500:
            if camcol == 4:
                dark_variance = 12.6025
            if camcol == 5:
                dark_variance = 2.1025

        extra_err = 1.0857 * np.sqrt(dark_variance + flux / gain) / flux
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

    elif survey == "GALEX":
        # https://asd.gsfc.nasa.gov/archive/galex/FAQ/counts_background.html
        CPS = flux
        if filt == "FUV":
            uv_err = -2.5 * (
                np.log10(CPS)
                - np.log10(
                    CPS
                    + (CPS * exptime + (0.050 * CPS * exptime) ** 2) ** 0.5 / exptime
                )
            )
            # flux_uv_err = (
            #    np.sqrt(CPS * exptime + (0.050 * CPS * exptime) ** 2) / exptime
            # )
        elif filt == "NUV":
            uv_err = -2.5 * (
                np.log10(CPS)
                - np.log10(
                    CPS
                    + (CPS * exptime + (0.027 * CPS * exptime) ** 2) ** 0.5 / exptime
                )
            )
            # flux_uv_err = (
            #    np.sqrt(CPS * exptime + (0.027 * CPS * exptime) ** 2) / exptime
            # )

        mag_err = np.sqrt(mag_err**2 + uv_err**2)

    elif survey == "2MASS":
        # see: https://wise2.ipac.caltech.edu/staff/jarrett/2mass/3chan/noise/
        S = flux
        N_c = 6  # number of coadd pixels
        k_z = 1.7  # kernel smoothing factor
        gain = 10  # typical gain in e-/dn
        n_f = ap_area  # number of frame pixels in the aperture; aprox. as aperture area
        n_c = 4 * n_f  # number of coadd pixels in the aperture
        sigma_c = bkg_rms  # coadd noise; assumed to be ~background noise

        SNR = S / np.sqrt(
            S / (gain * N_c)
            + 4 * n_c * k_z ** 2 * sigma_c ** 2
        )
        mag_err = 1.0857 / SNR


    elif "WISE" in survey:
        # see: https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec2_3f.html
        # see Table 5 of
        # https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4c.html#wpro
        # correction assumed to be 0 mags as PSF fitting is not used.
        m_apcor = 0.0
        f_apcor = 10 ** (-0.4 * m_apcor)
        F_src = f_apcor * flux

        N_p = ap_area  # effective number of noise-pixels characterizing the PRF
        # ratio of input (detector) pixel scale to output (Atlas Image) pixel scale
        pixel_scale_ratios = {"W1": 2, "W2": 2, "W3": 2, "W4": 4}
        Sin_Sout = pixel_scale_ratios[filt]
        F_corr = N_p * Sin_Sout**2
        
        k = np.pi / 2
        N_A = ap_area
        N_B = N_A / 10
        sigma_conf = bkg_rms  # assumed to be ~bkg
        sigma_src = np.sqrt(
            f_apcor ** 2 * F_corr * (bkg_rms ** 2 + k * (N_A ** 2) / N_B * bkg_rms ** 2)
            + sigma_conf ** 2
        )
        
        mag_err = np.sqrt(1.179 * sigma_src**2 / F_src**2)

        # add uncertainty from the ZP
        if survey == "unWISE":
            # These values are the same for all Atlas Images of a given band...
            # see: https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec2_3f.html
            unc_dict = {"W1": 0.006, "W2": 0.007, "W3": 0.012, "W4": 0.012}
            zp_unc = unc_dict[filt]
        elif survey == "WISE":
            zp_unc = header["MAGZPUNC"]

        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    elif survey == "LegacySurvey":
        # photometry uncertainties for DR10 from https://ui.adsabs.harvard.edu/abs/2023RNAAS...7..105Z/abstract
        # LS also includes uncertainties from inverse-variance maps, calculated together with the flux
        # (outside this function), as mentioned by Dustin Lang
        unc_dict = {
            "g": 5.0e-3,
            "r": 3.9e-3,
            "i": 4.3e-3,
            "z": 5.5e-3,
        }
        extra_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

    elif survey == "Spitzer":
        # already added at the beginning
        pass
    elif survey in ["VISTA", "UKIDSS"]:
        # add uncertainty from the ZP
        zp_unc = header["MAGZRR"]
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    elif survey == "HST":
        zp_unc = get_HST_err(filt, header)
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    elif survey == "SkyMapper":
        zp_unc = header["ZPTERR"]
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    elif survey == "SPLUS":
        # following Section 4.4 of Almeida-Fernandes et al. (2022)
        zp_uncs = {
            "U": 25e-3,
            "F395": 25e-3,
            "F378": 15e-3,
        }
        if filt in zp_uncs.keys():
            zp_unc = zp_uncs[filt]
        else:
            zp_unc = 1e-3
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    elif survey == "JWST":
        # see https://jwst-docs.stsci.edu/jwst-calibration-status/nircam-calibration-status/nircam-imaging-calibration-status#NIRCamImagingCalibrationStatus-Photometriccalibration
        # absolute flux calibration uncertainties < 1%, but set to 1% error for now
        zp_unc = 0.01 * (header["MAGZP"])
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    else:
        raise Exception(f"Survey {survey} has not been added for error propagation.")

    flux_err = np.abs(flux * 0.4 * np.log(10) * mag_err)

    return flux_err


def magnitude_calculation(
    flux: float,
    flux_err: float,
    survey: str,
    filt: str = None,
    ap_area: float = 0.0,
    header: Optional[fits.Header] = None,
    bkg_rms: float = 0.0,
) -> tuple[float, float, float, float, float]:
    """Calculates the calibrated magnitudes and errors.

    Error propagation is included here, both for magnitudes
    and fluxes.

    Parameters
    ----------
    flux: Aperture flux.
    survey: Survey name: e.g. ``PanSTARRS``, ``GALEX``.
    filt: Survey-specific filter.
    ap_area: Aperture area.
    header: Header of an image.
    bkg_rms: Background noise.

    Returns
    -------
    mag: Apparent magnitude.
    mag_err: Apparent magnitude uncertainty.
    flux: Total flux.
    flux_err: Total uncertainty in flux.
    zp: Zeropoint.
    """
    # get zeropoint
    zp = survey_zp(survey, filt)
    if zp == "header":
        zp = header["MAGZP"]

    if survey == "SDSS":
        # SDSS zero-points are not exactly in AB:
        # https://www.sdss4.org/dr12/algorithms/fluxcal/#SDSStoAB
        if filt == "u":
            offset = -0.04  # mag
        elif filt == "z":
            offset = 0.02  # mag
        else:
            offset = 0
        flux *= 10 ** (-0.4 * offset)
        flux_err *= 10 ** (-0.4 * offset)

    if survey == "unWISE":
        # To improve the agreement between unWISE and AllWISE fluxes,
        # these offsets need to be applied: https://catalog.unwise.me/catalogs.html#absolute
        if filt == "W1":
            offset = -4e-3  # 4 mmag
        elif filt == "W2":
            offset = -32e-3  # 32 mmag
        else:
            offset = 0
        flux *= 10 ** (-0.4 * offset)
        flux_err *= 10 ** (-0.4 * offset)

    if survey == "HST":
        # HST needs and aperture correction for the flux
        # see, e.g. https://www.stsci.edu/hst/instrumentation/acs/data-analysis/aperture-corrections
        ap_corr = correct_HST_aperture(filt, ap_area, header)
        flux = flux * ap_corr

    # error propagation
    flux_err = uncertainty_calculation(
        flux,
        flux_err,
        survey,
        filt,
        ap_area,
        header,
        bkg_rms,
    )

    pixel_units = survey_pixel_units(survey, filt)
    if pixel_units == "counts" and survey!="2MASS":
        # flux needs to be in units of counts per second
        # but only after the error propagation
        exptime = get_image_exptime(header, survey)
        flux /= exptime
        flux_err /= exptime

    mag = -2.5 * np.log10(flux) + zp
    mag_err = np.abs(2.5 * flux_err / (flux * np.log(10)))

    return mag, mag_err, flux, flux_err, zp


def extract_filter(
    filt: str, survey: str, version: Optional[str] = None
) -> tuple[np.ndarray, np.ndarray]:
    """Extracts the transmission function for the filter.

    Parameters
    ----------
    filt: Filter to extract.
    survey: Survey of the filters.
    version: Version of the filters to use. E.g. for the
        Legacy Survey as it uses DECam for the south
        and BASS+MzLS for the north.

    Returns
    -------
    wave: Wavelength range of the filter.
    transmission: Transmission function.
    """
    check_survey_validity(survey)
    if survey in ["HST", "JWST"]:
        check_filters_validity(filt, survey)
    else:
        check_filters_validity([filt], survey)

    if "WISE" in survey:
        survey = "WISE"  # for unWISE to use the same filters as WISE

    filters_path = hostphot_path.joinpath("filters", survey)

    # Assume DECaLS filters below 32 degrees and BASS+MzLS above
    # https://www.legacysurvey.org/status/
    if survey == "LegacySurvey":
        if version == "BASS+MzLS":
            if filt == "z":
                filt_file = filters_path / "MzLS_z.dat"
            else:
                filt_file = filters_path / f"BASS_{filt}.dat"
        elif version == "DECam":
            filt_file = filters_path / f"DECAM_{filt}.dat"
    elif survey == "HST":
        detector = filt.split("_")[1]
        if detector == "UVIS":
            # Usually UVIS2 is used, but there is no large difference
            filt_ = filt.replace("UVIS", "UVIS2")
        else:
            filt_ = filt
        filt_file = [file for file in filters_path.rglob(f"*{filt_}*")][0]
    elif survey == "JWST":
        filt_file = filters_path / f"{filt}.dat"
    else:
        filt_file = filters_path / f"{survey}_{filt}.dat"
    wave, transmission = np.loadtxt(filt_file).T

    return wave, transmission


def integrate_filter(
    spectrum_wave: np.ndarray,
    spectrum_flux: np.ndarray,
    filter_wave: np.ndarray,
    filter_response: np.ndarray,
    response_type: str = "photon",
) -> float:
    """Calcultes the flux density of an SED given a filter response.

    Parameters
    ----------
    spectrum_wave: Spectrum's wavelength range.
    spectrum_flux: Spectrum's flux density distribution.
    filter_wave: Filter's wavelength range.
    filter_response: Filter's response function.
    response_type: Filter's response type. Either ``photon`` or ``energy``.

    Returns
    -------
    flux_filter: Flux density.
    """
    assert response_type in ["energy", "photon"], "Not a valid response type!"
    if response_type == "energy":
        filter_response = filter_response.copy() / filter_wave

    interp_response = np.interp(
        spectrum_wave, filter_wave, filter_response, left=0.0, right=0.0
    )
    int1 = np.trapezoid(spectrum_flux * interp_response * spectrum_wave, spectrum_wave)
    int2 = np.trapezoid(filter_response * filter_wave, filter_wave)
    flux_filter = int1 / int2

    return flux_filter



import numpy as np
from typing import Optional
import matplotlib.pyplot as plt
import aplpy

from astropy import wcs
from astropy.io import fits
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.gaia import Gaia
from astroquery.mast import Catalogs

import sep
from hostphot._constants import font_family
from hostphot.utils import suppress_stdout

import warnings
from astropy.utils.exceptions import AstropyWarning

plt.rcParams["mathtext.fontset"] = "cm"


def extract_objects(
    data: np.ndarray,
    bkg: np.ndarray,
    host_ra: float,
    host_dec: float,
    threshold: float,
    img_wcs: wcs.WCS,
    dist_thresh: float = -1,
    deblend_cont: float = 0.005,
) -> tuple[np.ndarray, np.ndarray]:
    """Extracts objects and their ellipse parameters. The function :func:`sep.extract()`
    is used.

    If there is no detected object within a distance of ``dist_thresh`` from the galaxy
    coordinates, it means that the galaxy was not correctly identified.

    Parameters
    ----------
    data: Image data.
    bkg: Background level of the image.
    host_ra: Host-galaxy Right ascension of the galaxy in degrees.
    host_dec: Host-galaxy Declination of the galaxy in degrees.
    threshold: Source with flux above ``threshold*bkg_rms`` are extracted.
        See :func:`sep.extract()` for more information.
    img_wcs: Image's WCS.
    pixel_scale: Pixel scale, in units of arcsec/pixel, used to convert from pixel units
        to arcseconds.
    dist_thresh: Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    deblend_cont : Minimum contrast ratio used for object deblending. Default is 0.005.
        To entirely disable deblending, set to 1.0.

    Returns
    -------
    gal_obj: Galaxy object extracted.
    nogal_objs: All objects extracted except for the galaxy.
    """
    # extract objects with Source Extractor
    objects = sep.extract(data, threshold, err=bkg, deblend_cont=deblend_cont)

    host_coords = SkyCoord(
        ra=host_ra, dec=host_dec, unit=(u.degree, u.degree), frame="icrs"
    )
    objs_coords = img_wcs.pixel_to_world(objects["x"], objects["y"])
    distances = host_coords.separation(objs_coords).to(u.arcsec)
    dist_arcsec = distances.value

    if dist_thresh <= 0.0:
        dist_thresh = np.inf

    if any(dist_arcsec <= dist_thresh):
        gal_id = np.argmin(dist_arcsec)
        gal_obj = objects[gal_id : gal_id + 1]
    else:
        gal_obj = None
        gal_id = -99
        print("WARNING: the galaxy was no detected")

    objs_id = [i for i in range(len(objects)) if i != gal_id]
    nogal_objs = objects.take(objs_id)
    return gal_obj, nogal_objs


def find_gaia_objects(ra: float, dec: float, rad: float = 0.15) -> SkyCoord:
    """Finds objects using the Gaia DR3 catalog for the given
    coordinates in a given radius.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    rad: Search radius in degrees.

    Returns
    -------
    gaia_coord: Coordinates of the objects found.
    """
    Gaia.MAIN_GAIA_TABLE = "gaiaedr3.gaia_source"
    Gaia.ROW_LIMIT = -1
    coord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    width = u.Quantity(rad, u.deg)
    height = u.Quantity(rad, u.deg)
    try:
        with suppress_stdout():
            gaia_cat = Gaia.query_object_async(
                coordinate=coord, width=width, height=height
            )
    except Exception as exc:
        print(exc)
        print("No objects found with Gaia DR3, switching to DR2")
        Gaia.MAIN_GAIA_TABLE = "gaiadr2.gaia_source"
        gaia_cat = Gaia.query_object_async(coordinate=coord, width=width, height=height)

    gaia_ra = np.array(gaia_cat["ra"].value)
    gaia_dec = np.array(gaia_cat["dec"].value)
    gaia_coord = SkyCoord(
        ra=gaia_ra, dec=gaia_dec, unit=(u.degree, u.degree), frame="icrs"
    )
    return gaia_coord


def find_catalog_objects(ra: float, dec: float, rad: float = 0.15) -> SkyCoord:
    """Finds objects using the TESS image cutouts (Tic) catalog
    for the given coordinates in a given radius.

    **Note:** this catalog includes objects from these sources:
        HIP, TYC, UCAC, TWOMASS, SDSS, ALLWISE, GAIA, APASS, KIC

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    rad: Search radius in degrees.

    Returns
    -------
    cat_coord: Coordinates of the objects found.
    """
    coord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    cat_data = Catalogs.query_criteria(
        catalog="Tic", radius=rad, coordinates=coord, objType="STAR"
    )

    cat_ra = np.array(cat_data["ra"].value)
    cat_dec = np.array(cat_data["dec"].value)
    cat_coord = SkyCoord(
        ra=cat_ra, dec=cat_dec, unit=(u.degree, u.degree), frame="icrs"
    )
    return cat_coord


def cross_match(
    objects: np.ndarray, img_wcs: wcs.WCS, coord: SkyCoord, dist_thresh: float = 1.0
) -> None:
    """Cross-matches objects with a given set of coordinates.
    Those with a distance of less than ``dist_thresh`` are selected.

    Parameters
    ----------
    objects: Objects detected with :func:`sep.extract()`.
    img_wcs: WCS of the image from which the objects were extracted.
    coord: Coordinates for the cross-match.
    dist_thresh: Distance in arcsec to crossmatch the objects with
        the given coordinates.
    """
    # coordinates in arcsec
    objs_coord = img_wcs.pixel_to_world(objects["x"], objects["y"])
    objs_ra = objs_coord.ra.to(u.arcsec).value
    objs_dec = objs_coord.dec.to(u.arcsec).value

    cat_ra = np.array(coord.ra.to(u.arcsec).value)
    cat_dec = np.array(coord.dec.to(u.arcsec).value)

    objs_id = []
    for i in range(len(objects)):
        ra_dist = objs_ra[i] - cat_ra
        dec_dist = objs_dec[i] - cat_dec
        dist_arcsec = np.sqrt(ra_dist**2 + dec_dist**2)
        if any(dist_arcsec <= dist_thresh):
            objs_id.append(i)

    objs = objects.take(objs_id)

    return objs


def plot_detected_objects(
    hdu: list[fits.ImageHDU],
    objects: np.ndarray,
    scale: float,
    ra: Optional[float] = None,
    dec: Optional[float] = None,
    host_ra: Optional[float] = None,
    host_dec: Optional[float] = None,
    title: Optional[str] = None,
    outfile: Optional[str] = None,
) -> None:
    """Plots the objects extracted with :func:`sep.extract()``.

    Parameters
    ----------
    hdu: ~astropy.io.fits
        Header Data Unit.
    objects: array
        Objects detected with :func:`sep.extract()`.
    scale: float
        Scale of the ellipse's semi-mayor and semi-minor axes.
    img_wcs: WCS
        Image's WCS.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    host_ra: float, default ``None``
       Right ascension of a galaxy.
    host_dec: float, default ``None``
       Declination of a galaxy.
    title: str, default ``None``
        Title of the image.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """
    figure = plt.figure(figsize=(10, 10))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig = aplpy.FITSFigure(hdu, figure=figure)

    with suppress_stdout():
        fig.show_grayscale(stretch="arcsinh")

    # plot SN marker
    if (ra is not None) and (dec is not None):
        fig.show_markers(
            ra,
            dec,
            edgecolor="k",
            facecolor="aqua",
            marker="*",
            s=250,
            label="SN",
        )

    # plot galaxy marker and aperture
    if (host_ra is not None) and (host_dec is not None):
        fig.show_markers(
            host_ra,
            host_dec,
            edgecolor="k",
            facecolor="m",
            alpha=0.7,
            marker="P",
            s=250,
            label="Galaxy position",
        )
    fig.show_ellipses(
        objects["x"],
        objects["y"],
        2 * scale * objects["a"],
        2 * scale * objects["b"],
        objects["theta"][0] * 180.0 / np.pi,
        coords_frame="pixel",
        linewidth=3,
        edgecolor="r",
    )
    # ticks
    fig.tick_labels.set_font(**{"family": font_family, "size": 18})
    fig.tick_labels.set_xformat("dd.dd")
    fig.tick_labels.set_yformat("dd.dd")
    fig.ticks.set_length(6)
    # other configuration options
    # ToDo: solve this deprecation warning (Aplpy should do it?)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig.axis_labels.set_font(**{"family": font_family, "size": 18})
    fig.set_title(title, **{"family": font_family, "size": 24})
    fig.set_theme("publication")
    fig.ax.legend(fancybox=True, framealpha=1, prop={"size": 20, "family": font_family})
    if outfile:
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(figure)
    else:
        plt.show()



import importlib
import pandas as pd
from pathlib import Path
import astropy.units as u
from typing import Optional

from hostphot._constants import workdir
from hostphot.utils import check_work_dir, store_input
from hostphot.surveys_utils import (
    get_survey_filters,
    check_survey_validity,
)


def download_images(
    name: str,
    ra: float,
    dec: float,
    survey: str,
    size: float | u.Quantity = 3,
    filters: Optional[str | list] = None,
    overwrite: bool = True,
    version: str = None,
    save_input: bool = True,
) -> None:
    """Download images for a given object in the given filters of a
    given survey.

    The surveys that use the ``version`` parameter are: GALEX (``AIS``, ``MIS``,
    ``DIS``, ``NGS`` and ``GII``),  unWISE (``allwise`` and ``neo{i}`` for {i}=1-7),
    VISTA (``VHS``, ``VIDEO`` and ``VIKING``), SDSS (``dr{i}`` for {i}=12-17)
    and LegacySurvey (``dr{i}`` for {i}=?-10).

    Parameters
    ----------
    name: Name used for tracking the object in your localdirectory.
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    survey: Survey used to download the images.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters for the images.
    overwrite: If ``True``, the images are overwritten if they alreadyexist.
    version: Version used by some surveys including multiple surveys. E.g. ``VHS`` for VISTA.
    save_input: Whether to save the input parameters.

    Examples
    --------
    >>> from hostphot.cutouts import download_images
    >>> name = 'SN2004eo'
    >>> host_ra, host_dec = 308.2092, 9.92755  # coords of host galaxy of SN2004eo
    >>> survey = 'PanSTARRS'
    >>> download_images(name, host_ra, host_dec, survey=survey)
    """
    input_params = locals()  # dictionary
    # initial checks
    check_survey_validity(survey)
    check_work_dir(workdir)
    survey_dir = Path(workdir, name, survey)
    # check output directory and filters
    if survey_dir.is_dir() is False:
        survey_dir.mkdir(parents=True)
    if filters is None:
        filters = get_survey_filters(survey)

    # save input parameters
    if save_input is True:
        inputs_file = survey_dir / "input_cutouts.csv"
        store_input(input_params, inputs_file)

    # check existing images
    if overwrite is False:
        filters_without_image = []
        for filt in filters:
            if survey == "HST":
                inst = version.replace("/", "_")
                filt_image = survey_dir / f"{survey}_{inst}_{filters}.fits"
            else:
                filt_image = survey_dir / f"{survey}_{filt}.fits"
            if filt_image.is_file() is False:
                filters_without_image.append(filt)
        # only download images not found locally
        filters = filters_without_image
        if len(filters) == 0:
            return None

    # extract download function for the given survey
    if survey == "2MASS":
        survey_module = importlib.import_module("hostphot.cutouts.twomass")
    elif survey == "unWISE":
        survey_module = importlib.import_module("hostphot.cutouts.wise")
    else:
        survey_module = importlib.import_module(f"hostphot.cutouts.{survey.lower()}")

    # download the images
    get_images = getattr(survey_module, f"get_{survey}_images")  # this is a function
    if survey in ["SDSS", "GALEX", "unWISE", "LegacySurvey"]:
        hdu_list = get_images(ra, dec, size, filters, version)
    else:
        hdu_list = get_images(ra, dec, size, filters)
    if hdu_list is None:
        return None
    # save the images
    for hdu, filt in zip(hdu_list, filters):
        if hdu is None:
            continue  # skip missing filter/image
        # get output file name
        if survey == "HST":
            #inst = version.replace("/", "_")
            #outfile = Path(survey_dir, f"{survey}_{inst}_{filters}.fits")
            outfile = Path(survey_dir, f"{survey}_{filt}.fits")
        else:
            outfile = Path(survey_dir, f"{survey}_{filt}.fits")
        if overwrite is True or outfile.is_file() is False:
            hdu.writeto(outfile, overwrite=overwrite)
            hdu.close()
        else:
            hdu.close()

    # remove directory if it remains empty
    try:
        survey_dir.rmdir()
    except:
        pass



import numpy as np
import pandas as pd
from pathlib import Path
from typing import Optional

from pyvo.dal import sia
import astropy.units as u
from astropy.io import fits

import hostphot
hostphot_path = Path(hostphot.__path__[0])
from hostphot.surveys_utils import get_survey_filters, check_filters_validity

def get_SPLUS_urls(ra: float, dec: float, fov: float | u.Quantity = 3, 
                    filters: Optional[str | list] = None) -> list[str] | None:
    """Obtains the URLs of the S-Plus images.
    
    The available filters are: 'F378', 'F395', 'F410', 'F430', 
    'F515', 'F660', 'F861', 'G', 'I', 'R', 'U', 'Z'.

    Parameters
    ----------
    ra: Right ascension in degrees.
    dec: Declination in degrees.
    fov: Field of view in degrees.
    filters: Filters to use. If ``None``, uses all available filters.

    Returns
    -------
    url_list: List of URLs with S-Plus images.
    """
    splus_url = "https://datalab.noirlab.edu/sia/splus_dr1"
    svc = sia.SIAService(splus_url)
    imgs_table = svc.search(
        (ra, dec), (fov / np.cos(dec * np.pi / 180), fov), verbosity=2
    )
    if len(imgs_table) == 0:
        print(("Warning: empty table returned for " f"ra={ra}, dec={dec}"))
        return None

    imgs_df = pd.DataFrame(imgs_table)
    imgs_df = imgs_df[imgs_df.access_format.str.endswith('fits')]
    if len(imgs_df) == 0:
        return None
    url_list = []
    for filt in filters:
        filt_df = imgs_df[imgs_df.obs_bandpass==filt]
        if len(filt_df) == 0:
            url_list.append(None)
        else:
            fits_url = filt_df.access_url.values[0]  # first image
            url_list.append(fits_url)    
    return url_list

def get_SPLUS_images(ra: float, dec: float, size: float | u.Quantity = 3, 
                    filters: Optional[str] = None) -> fits.HDUList | None:
    """Gets S-Plus fits images for the given coordinates and
    filters.

    The available filters are: 'F378', 'F395', 'F410', 'F430', 
    'F515', 'F660', 'F861', 'G', 'I', 'R', 'U', 'Z'.

    Parameters
    ----------
    ra: Right Ascension in degrees.
    dec: Declination in degrees.
    size: Image size. If a float is given, the units are assumed to be arcmin.
    filters: Filters to use. If ``None``, uses all available filters.

    Returns
    -------
    hdu_list: List of fits images.
    """
    survey = "SPLUS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    if isinstance(size, (float, int)):
        fov = (size * u.arcmin).to(u.degree).value
    else:
        fov = size.to(u.degree).value

    url_list = get_SPLUS_urls(ra, dec, fov, filters)
    if url_list is None:
        return None
    hdu_list = []
    for url in url_list:
        if url is None:
            hdu_list.append(None)
        else:
            hdu = fits.open(url)
            # add zeropoint
            # file from https://splus.cloud/documentation/dr2_3
            zps_file = hostphot_path.joinpath('filters', 'SPLUS', 'iDR3_zps.cat')
            zps_df = pd.read_csv(zps_file, sep='\\s+')
            field = hdu[0].header['OBJECT'].replace('_', '-')
            field_df = zps_df[zps_df['#field']==field]
            img_filt = hdu[0].header['FILTER']
            zp = field_df[img_filt].values[0]
            hdu[0].header['MAGZP'] = zp
            # initial EXPTIME is normalised, so it doesn't help
            hdu[0].header['EXPTIME'] = hdu[0].header['TEXPOSED']
            hdu_list.append(hdu)
    return hdu_list


import os
import sys
import pandas as pd
from pathlib import Path
import matplotlib.image as img
import matplotlib.pyplot as plt
from contextlib import contextmanager
from datetime import datetime, timezone

import aplpy
from astropy.io import fits
from hostphot._constants import font_family

import warnings
from astropy.utils.exceptions import AstropyWarning

def check_work_dir(wokrdir: str | Path) -> None:
    """Checks if the working directory exists. If it
    does not, one is created.

    Parameters
    ----------
    wokrdir: str
        Working directory path.
    """
    work_path = Path(wokrdir)
    if work_path.is_dir() is False:
        work_path.mkdir(parents=True)


@contextmanager
def suppress_stdout():
    """Suppresses annoying outputs.

    Useful with astroquery and aplpy packages.
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            
def store_input(input_params: dict, inputs_file: Path) -> None:
    """Stores the input parameters of a function.

    Parameters
    ----------
    inputs_params: the parameters of a function.
    inputs_file: where to store the parameters.
    """
    # save input parameters
    now = datetime.now(timezone.utc).isoformat()
    input_params.update({"timestamp": now})
    inputs_df = pd.DataFrame([input_params])
    if inputs_file.exists():
        inputs_df.to_csv(inputs_file, index=False, mode="a", header=False)
    else:
        inputs_df.to_csv(inputs_file, index=False)
        
            
def plot_fits(fits_file: str | Path | list[fits.ImageHDU], ext: int = 0) -> None:
    """Plots a FITS file.

    Parameters
    ----------
    fits_file: FITS file.
    ext: Extension index.
    """
    if isinstance(fits_file, str) or isinstance(fits_file, Path):
        title = Path(fits_file).name.split()[0]
        fits_file = str(fits_file)
    else:
        title = None

    figure = plt.figure(figsize=(10, 10))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig = aplpy.FITSFigure(fits_file, hdu=ext, figure=figure)
    with suppress_stdout():
        fig.show_grayscale(stretch="arcsinh")

    # ticks
    fig.tick_labels.set_font(**{"family": font_family, "size": 18})
    fig.tick_labels.set_xformat("dd.dd")
    fig.tick_labels.set_yformat("dd.dd")
    fig.ticks.set_length(6)
    # ToDo: solve this deprecation warning (Aplpy should do it?)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig.axis_labels.set_font(**{"family": font_family, "size": 18})
    # title + theme
    fig.set_title(title, **{"family": font_family, "size": 24})
    fig.set_theme("publication")
    plt.show()


def plot_image(image_file: str | Path) -> None:
    """Plots an image file.

    E.g. '.png' and '.jpg' files.

    Parameters
    ----------
    image_file: Path to an image file.
    """
    _, ax = plt.subplots(figsize=(16, 8))
    im = img.imread(image_file)
    ax.imshow(im)
    plt.axis("off")
    plt.show()



# -*- coding: utf-8 -*-

from .coadd import coadd_images
#from objects_detection import extract_objects
from .masking import create_mask
from .objects_detection import extract_objects, plot_detected_objects